# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import os
import json
import torch
import ntpath
import shutil
import onnxruntime as ort
from PIL import Image as PilImage
from torchvision.transforms import transforms as T
from opendr.engine.learners import Learner
from opendr.engine.datasets import DatasetIterator, ExternalDataset, MappedDatasetIterator
from opendr.perception.object_tracking_2d.logger import Logger
from opendr.perception.object_tracking_2d.datasets.market1501_dataset import Market1501DatasetIterator
from opendr.perception.object_tracking_2d.deep_sort.algorithm.run import train
from opendr.perception.object_tracking_2d.fair_mot.algorithm.run import evaluate
from opendr.perception.object_tracking_2d.deep_sort.algorithm.deep_sort_tracker import DeepSortTracker
from opendr.engine.data import Image, ImageWithDetections
from opendr.engine.constants import OPENDR_SERVER_URL
from urllib.request import urlretrieve


class ObjectTracking2DDeepSortLearner(Learner):
    def __init__(
        self,
        lr=0.1,
        iters=-1,
        batch_size=4,
        optimizer="sgd",
        checkpoint_after_iter=0,
        checkpoint_load_iter=0,
        temp_path="",
        device="cuda",
        max_dist=0.2,
        min_confidence=0.3,
        nms_max_overlap=0.5,
        max_iou_distance=0.7,
        max_age=70,
        n_init=3,
        nn_budget=100,
    ):
        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        super(ObjectTracking2DDeepSortLearner, self).__init__(
            lr=lr,
            iters=iters,
            batch_size=batch_size,
            optimizer=optimizer,
            checkpoint_after_iter=checkpoint_after_iter,
            checkpoint_load_iter=checkpoint_load_iter,
            temp_path=temp_path,
            device=device,
        )

        self.max_dist = max_dist
        self.min_confidence = min_confidence
        self.nms_max_overlap = nms_max_overlap
        self.max_iou_distance = max_iou_distance
        self.max_age = max_age
        self.n_init = n_init
        self.nn_budget = nn_budget

        self.__create_model()
        self.model_optimizer = torch.optim.SGD(
            self.tracker.deepsort.extractor.net.parameters(), lr, momentum=0.9, weight_decay=5e-4
        )

        self.infers_count = 0
        self.infers_time = 0

    def save(self, path, verbose=False):
        """
        This method is used to save a trained model.
        Provided with the path, absolute or relative, including a *folder* name, it creates a directory with the name
        of the *folder* provided and saves the model inside with a proper format and a .json file with metadata.
        If self.optimize was ran previously, it saves the optimized ONNX model in a similar fashion, by copying it
        from the self.temp_path it was saved previously during conversion.
        :param path: for the model to be saved, including the folder name
        :type path: str
        :param verbose: whether to print success message or not, defaults to 'False'
        :type verbose: bool, optional
        """

        if self.tracker is None:
            raise UserWarning("No model is loaded, cannot save.")

        folder_name, _, tail = self.__extract_trailing(path)  # Extract trailing folder name from path
        # Also extract folder name without any extension if extension is erroneously provided
        folder_name_no_ext = folder_name.split(sep='.')[0]

        # Extract path without folder name, by removing folder name from original path
        path_no_folder_name = ''.join(path.rsplit(folder_name, 1))
        # If tail is '', then path was a/b/c/, which leaves a trailing double '/'
        if tail == '':
            path_no_folder_name = path_no_folder_name[0:-1]  # Remove one '/'

        # Create model directory
        new_path = path_no_folder_name + folder_name_no_ext
        os.makedirs(new_path, exist_ok=True)

        model_metadata = {"model_paths": [], "framework": "pytorch", "format": "", "has_data": False,
                          "inference_params": {}, "optimized": None, "optimizer_info": {}}

        if self.tracker.deepsort.extractor.net.ort_session is None:
            model_metadata["model_paths"] = [
                folder_name_no_ext + ".pth",
            ]
            model_metadata["optimized"] = False
            model_metadata["format"] = "pth"

            torch.save({
                'state_dict': self.tracker.deepsort.extractor.net.state_dict()
            }, os.path.join(path_no_folder_name, folder_name_no_ext, model_metadata["model_paths"][0]))
            if verbose:
                print("Saved Pytorch model.")
        else:
            model_metadata["model_paths"] = [
                folder_name_no_ext + ".onnx"
            ]
            model_metadata["optimized"] = True
            model_metadata["format"] = "onnx"

            shutil.copy2(
                os.path.join(self.temp_path, "onnx_model_temp.onnx"),
                os.path.join(path_no_folder_name, folder_name_no_ext, model_metadata["model_paths"][0])
            )
            if verbose:
                print("Saved ONNX model.")

        with open(os.path.join(new_path, folder_name_no_ext + ".json"), 'w') as outfile:
            json.dump(model_metadata, outfile)

    def load(
        self,
        path,
        verbose=False,
    ):
        """
        Loads the model from inside the path provided, based on the metadata .json file included.
        :param path: path of the directory the model was saved
        :type path: str
        :param verbose: whether to print success message or not, defaults to 'False'
        :type verbose: bool, optional
        """

        model_name, _, _ = self.__extract_trailing(path)  # Trailing folder name from the path provided

        with open(os.path.join(path, model_name + ".json")) as metadata_file:
            metadata = json.load(metadata_file)

        if not metadata["optimized"]:
            self.__load_from_pth(self.tracker.deepsort.extractor.net, os.path.join(path, metadata["model_paths"][0]))
            if verbose:
                print("Loaded Pytorch model.")
        else:
            self.__load_rpn_from_onnx(os.path.join(path, metadata["model_paths"][0]))
            if verbose:
                print("Loaded ONNX model.")

    def reset(self):
        self.tracker.reset()

    def fit(
        self,
        dataset,
        epochs,
        val_dataset=None,
        val_epochs=-1,
        logging_path=None,
        silent=False,
        verbose=False,
        train_transforms=None,
        val_transforms=None,
    ):

        logger = Logger(silent, verbose, logging_path)

        (
            input_dataset_iterator,
            eval_dataset_iterator,
        ) = self._prepare_datasets(
            dataset,
            val_dataset,
            require_val_dataset=val_epochs > 0,
        )

        checkpoints_path = os.path.join(self.temp_path, "checkpoints")
        if self.checkpoint_after_iter != 0 or self.checkpoint_load_iter != 0:
            os.makedirs(checkpoints_path, exist_ok=True)

        if self.checkpoint_load_iter != 0:
            state_dict = torch.load(f"checkpoint_{self.checkpoint_load_iter}.pth")
            self.tracker.deepsort.extractor.net.load_state_dict(state_dict)

        if train_transforms is None:
            train_transforms = T.Compose(
                [
                    T.RandomCrop((128, 64), padding=4),
                    T.RandomHorizontalFlip(),
                    T.ToTensor(),
                    T.Normalize(
                        [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
                    ),
                ]
            )

        if val_transforms is None:
            val_transforms = T.Compose(
                [
                    T.Resize((128, 64)),
                    T.ToTensor(),
                    T.Normalize(
                        [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
                    ),
                ]
            )

        log_interval = 1 if verbose else 50

        result = train(
            self.tracker.deepsort.extractor.net, input_dataset_iterator, epochs,
            self.iters, eval_dataset_iterator, self.model_optimizer,
            self.batch_size, val_epochs, train_transforms, val_transforms,
            self.device, logger.log, log_interval, checkpoints_path, self.checkpoint_after_iter
        )

        logger.close()

        return result

    def eval(
        self,
        dataset,
        logging_path=None,
        silent=False,
        verbose=False,
    ):

        logger = Logger(silent, verbose, logging_path)

        eval_dataset_iterator = self._prepare_eval_dataset(
            dataset,
        )

        result = evaluate(self.infer, eval_dataset_iterator)

        logger.log(Logger.LOG_WHEN_NORMAL, result)

        logger.close()

        return result

    def infer(self, batch, frame_ids=None, swap_left_top=False):

        if self.tracker is None:
            raise ValueError("No model loaded or created")

        is_single_image = False

        if isinstance(batch, ImageWithDetections):
            batch = [batch]
            is_single_image = True
        elif not isinstance(batch, list):
            raise ValueError("Input batch should be an engine.ImageWithDetections or a list of engine.ImageWithDetections")

        if frame_ids is None:
            frame_ids = [-1] * len(batch)
        elif is_single_image:
            frame_ids = [frame_ids]

        results = []

        for image, frame_id in zip(batch, frame_ids):

            t0 = time.time()

            result = self.tracker.infer(image, frame_id, swap_left_top=swap_left_top)
            results.append(result)

            t0 = time.time() - t0
            self.infers_count += 1
            self.infers_time += t0

        if is_single_image:
            results = results[0]

        return results

    def optimize(self, do_constant_folding=False, img_size=(64, 128)):
        """
        Optimize method converts the model to ONNX format and saves the
        model in the parent directory defined by self.temp_path. The ONNX model is then loaded.
        :param do_constant_folding: whether to optimize constants, defaults to 'False'
        :type do_constant_folding: bool, optional
        """

        if self.tracker.deepsort.extractor.net is None:
            raise UserWarning("No model is loaded, cannot optimize. Load or train a model first.")
        if self.tracker.deepsort.extractor.net.ort_session is not None:
            raise UserWarning("Model is already optimized in ONNX.")

        input_shape = [
            1,
            3,
            img_size[1],
            img_size[0],
        ]

        try:
            self.__convert_to_onnx(
                input_shape,
                os.path.join(self.temp_path, "onnx_model_temp.onnx"), do_constant_folding
            )
        except FileNotFoundError:
            # Create temp directory
            os.makedirs(self.temp_path, exist_ok=True)
            self.__convert_to_onnx(
                input_shape,
                os.path.join(self.temp_path, "onnx_model_temp.onnx"), do_constant_folding
            )

        self.__load_from_onnx(os.path.join(self.temp_path, "onnx_model_temp.onnx"))

    @staticmethod
    def download(model_name, path, server_url=None):

        if server_url is None and model_name not in [
            "deep_sort",
        ]:
            raise ValueError("Unknown model_name: " + model_name)

        os.makedirs(path, exist_ok=True)

        if server_url is None:
            server_url = os.path.join(
                OPENDR_SERVER_URL, "perception", "object_tracking_2d",
                "deep_sort"
            )

        url = os.path.join(
            server_url, model_name
        )

        model_dir = os.path.join(path, model_name)
        os.makedirs(model_dir, exist_ok=True)

        urlretrieve(os.path.join(
            url, model_name + ".json"
        ), os.path.join(
            model_dir, model_name + ".json"
        ))

        try:
            urlretrieve(os.path.join(
                url, model_name + ".pth"
            ), os.path.join(
                model_dir, model_name + ".pth"
            ))
        except Exception:
            urlretrieve(os.path.join(
                url, model_name + ".tckpt"
            ), os.path.join(
                model_dir, model_name + ".pth"
            ))

        print("Downloaded model", model_name, "to", model_dir)

        return model_dir

    def __convert_to_onnx(self, input_shape, output_name, do_constant_folding=False, verbose=False):
        inp = torch.randn(input_shape).to(self.device)
        input_names = ["data"]
        output_names = ["output"]

        torch.onnx.export(
            self.tracker.deepsort.extractor.net, inp, output_name, verbose=verbose, opset_version=11,
            do_constant_folding=do_constant_folding, input_names=input_names, output_names=output_names,
            dynamic_axes={"data": {0: "batch"}},
        )

    def __load_from_onnx(self, path):
        """
        This method loads an ONNX model from the path provided into an onnxruntime inference session.

        :param path: path to ONNX model
        :type path: str
        """
        self.tracker.deepsort.extractor.net.ort_session = ort.InferenceSession(path)

        # The comments below are the alternative way to use the onnx model, it might be useful in the future
        # depending on how ONNX saving/loading will be implemented across the toolkit.
        # # Load the ONNX model
        # self.tracker.deepsort.extractor.net = onnx.load(path)
        #
        # # Check that the IR is well formed
        # onnx.checker.check_model(self.tracker.deepsort.extractor.net)
        #
        # # Print a human readable representation of the graph
        # onnx.helper.printable_graph(self.tracker.deepsort.extractor.net.graph)

    def __load_from_pth(self, model, path, use_original_dict=False):
        all_params = torch.load(path, map_location=self.device)
        model.load_state_dict(all_params if use_original_dict else all_params["state_dict"])

    def _prepare_datasets(
        self,
        dataset,
        val_dataset,
        require_dataset=True,
        require_val_dataset=True,
    ):

        input_dataset_iterator = None
        eval_dataset_iterator = None

        def map_dataset(dataset):

            def image_to_pil(image: Image):

                return PilImage.fromarray(image.convert('channels_last'))

            return MappedDatasetIterator(dataset, lambda data: (image_to_pil(data[0]), data[1].data))

        if isinstance(dataset, ExternalDataset):

            dataset_path = dataset.path
            if dataset.dataset_type.lower() != "market1501":
                raise ValueError(
                    "ExternalDataset (" + str(dataset) +
                    ") is given as a dataset, but it is not a Market1501 dataset")

            input_dataset_iterator = map_dataset(Market1501DatasetIterator(os.path.join(dataset_path, "bounding_box_train")))
        elif isinstance(dataset, DatasetIterator):
            input_dataset_iterator = map_dataset(dataset)
        else:
            if require_dataset or dataset is not None:
                raise ValueError(
                    "dataset parameter should be an ExternalDataset or a DatasetIterator"
                )

        if isinstance(val_dataset, ExternalDataset):

            val_dataset_path = val_dataset.path
            if val_dataset.dataset_type.lower() != "market1501":
                raise ValueError(
                    "ExternalDataset (" + str(val_dataset) +
                    ") is given as a val_dataset, but it is not a Market1501 dataset"
                )

            eval_dataset_iterator = map_dataset(Market1501DatasetIterator(os.path.join(val_dataset_path, "bounding_box_test")))

        elif isinstance(val_dataset, DatasetIterator):
            eval_dataset_iterator = map_dataset(val_dataset)
        elif val_dataset is None:
            if isinstance(dataset, ExternalDataset):
                val_dataset_path = dataset.path
                if dataset.dataset_type.lower() != "market1501":
                    raise ValueError(
                        "ExternalDataset (" + str(dataset) +
                        ") is given as a dataset, but it is not a Market1501 dataset"
                    )

                eval_dataset_iterator = map_dataset(Market1501DatasetIterator(
                    os.path.join(val_dataset_path, "bounding_box_test")
                ))

            elif require_val_dataset:
                raise ValueError(
                    "val_dataset is None and can't be derived from" +
                    " the dataset object because the dataset is not an ExternalDataset"
                )
            else:
                eval_dataset_iterator = map_dataset(input_dataset_iterator)
        else:
            raise ValueError(
                "val_dataset parameter should be an ExternalDataset or a DatasetIterator or None"
            )

        return input_dataset_iterator, eval_dataset_iterator

    def _prepare_eval_dataset(
        self,
        dataset,
    ):

        eval_dataset_iterator = None

        if isinstance(dataset, DatasetIterator):
            eval_dataset_iterator = dataset
        else:
            raise ValueError(
                "dataset parameter should be a DatasetIterator"
            )

        return eval_dataset_iterator

    def __create_model(self):

        self.tracker = DeepSortTracker(
            max_dist=self.max_dist,
            min_confidence=self.min_confidence,
            nms_max_overlap=self.nms_max_overlap,
            max_iou_distance=self.max_iou_distance,
            max_age=self.max_age,
            n_init=self.n_init,
            nn_budget=self.nn_budget,
            device=self.device,
        )

    @staticmethod
    def __extract_trailing(path):
        """
        Extracts the trailing folder name or filename from a path provided in an OS-generic way, also handling
        cases where the last trailing character is a separator. Returns the folder name and the split head and tail.
        :param path: the path to extract the trailing filename or folder name from
        :type path: str
        :return: the folder name, the head and tail of the path
        :rtype: tuple of three strings
        """
        head, tail = ntpath.split(path)
        folder_name = tail or ntpath.basename(head)  # handle both a/b/c and a/b/c/
        return folder_name, head, tail



import copy
import motmetrics as mm
import numpy as np


mm.lap.default_solver = "lap"


class Evaluator(object):
    def __init__(self):
        self.reset_accumulator()

    def reset_accumulator(self):
        self.acc = mm.MOTAccumulator(auto_id=True)

    def eval_frame(self, frame_id, gt_tlwhs, gt_ids, trk_tlwhs, trk_ids, rtn_events=False):
        # results
        trk_tlwhs = np.copy(trk_tlwhs)
        trk_ids = np.copy(trk_ids)

        iou_distance = mm.distances.iou_matrix(
            gt_tlwhs, trk_tlwhs, max_iou=0.5
        )

        # acc
        self.acc.update(gt_ids, trk_ids, iou_distance)

        if (
            rtn_events
            and iou_distance.size > 0
            and hasattr(self.acc, "last_mot_events")
        ):
            events = (
                self.acc.last_mot_events
            )  # only supported by https://github.com/longcw/py-motmetrics
        else:
            events = None
        return events

    def eval_all(self, gt_tlwhs, gt_ids, trk_tlwhs, trk_ids):
        self.reset_accumulator()

        for frame_id in range(len(gt_ids)):
            self.eval_frame(
                frame_id, gt_tlwhs[frame_id], gt_ids[frame_id],
                trk_tlwhs[frame_id], trk_ids[frame_id],
                rtn_events=False
            )

        return self.acc

    @staticmethod
    def get_summary(
        accs,
        names,
        metrics=(
            "mota",
            "num_switches",
            "idp",
            "idr",
            "idf1",
            "precision",
            "recall",
        ),
    ):
        names = copy.deepcopy(names)
        if metrics is None:
            metrics = mm.metrics.motchallenge_metrics
        metrics = copy.deepcopy(metrics)

        mh = mm.metrics.create()
        summary = mh.compute_many(
            accs, metrics=metrics, names=names, generate_overall=True
        )

        return summary

    @staticmethod
    def save_summary(summary, filename):
        import pandas as pd

        writer = pd.ExcelWriter(filename)
        summary.to_excel(writer)
        writer.save()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.models.losses import FocalLoss
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.models.losses import (
    RegL1Loss,
    RegLoss,
    NormRegL1Loss,
    RegWeightedL1Loss,
)
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.models.decode import (
    mot_decode,
)
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.models.utils import (
    _sigmoid,
    _tranpose_and_gather_feat,
)
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.utils.post_process import (
    ctdet_post_process,
)
from .base_trainer import BaseTrainer


class MotLoss(torch.nn.Module):
    def __init__(
        self,
        mse_loss,
        reg_loss,
        dense_wh,
        cat_spec_wh,
        reid_dim,
        nID,
        norm_wh,
        num_stacks,
        wh_weight,
        off_weight,
        id_weight,
        reg_offset,
        hm_weight,
    ):

        self.mse_loss = mse_loss
        self.reg_loss = reg_loss
        self.dense_wh = dense_wh
        self.cat_spec_wh = cat_spec_wh
        self.reid_dim = reid_dim
        self.nID = nID
        self.norm_wh = norm_wh
        self.num_stacks = num_stacks
        self.wh_weight = wh_weight
        self.off_weight = off_weight
        self.id_weight = id_weight
        self.reg_offset = reg_offset
        self.hm_weight = hm_weight

        super(MotLoss, self).__init__()
        self.crit = torch.nn.MSELoss() if mse_loss else FocalLoss()
        self.crit_reg = (
            RegL1Loss()
            if reg_loss == "l1"
            else RegLoss()
            if reg_loss == "sl1"
            else None
        )
        self.crit_wh = (
            torch.nn.L1Loss(reduction="sum")
            if dense_wh
            else NormRegL1Loss()
            if norm_wh
            else RegWeightedL1Loss()
            if cat_spec_wh
            else self.crit_reg
        )
        self.emb_dim = reid_dim
        self.nID = nID
        self.classifier = nn.Linear(self.emb_dim, self.nID)
        self.IDLoss = nn.CrossEntropyLoss(ignore_index=-1)
        self.emb_scale = math.sqrt(2) * math.log(self.nID - 1)
        self.s_det = nn.Parameter(-1.85 * torch.ones(1))
        self.s_id = nn.Parameter(-1.05 * torch.ones(1))

    def forward(self, outputs, batch):
        hm_loss, wh_loss, off_loss, id_loss = 0, 0, 0, 0
        for s in range(self.num_stacks):
            output = outputs[s]
            if not self.mse_loss:
                output["hm"] = _sigmoid(output["hm"])

            hm_loss += self.crit(output["hm"], batch["hm"]) / self.num_stacks
            if self.wh_weight > 0:
                wh_loss += (
                    self.crit_reg(
                        output["wh"], batch["reg_mask"], batch["ind"], batch["wh"],
                    )
                    / self.num_stacks
                )

            if self.reg_offset and self.off_weight > 0:
                off_loss += (
                    self.crit_reg(
                        output["reg"], batch["reg_mask"], batch["ind"], batch["reg"],
                    )
                    / self.num_stacks
                )

            if self.id_weight > 0:
                id_head = _tranpose_and_gather_feat(output["id"], batch["ind"])
                id_head = id_head[batch["reg_mask"] > 0].contiguous()
                id_head = self.emb_scale * F.normalize(id_head)
                id_target = batch["ids"][batch["reg_mask"] > 0]

                id_output = self.classifier(id_head).contiguous()
                id_loss += self.IDLoss(id_output, id_target)

        det_loss = (
            self.hm_weight * hm_loss
            + self.wh_weight * wh_loss
            + self.off_weight * off_loss
        )

        loss = (
            torch.exp(-self.s_det) * det_loss
            + torch.exp(-self.s_id) * id_loss
            + (self.s_det + self.s_id)
        )
        loss *= 0.5

        loss_stats = {
            "loss": loss,
            "hm_loss": hm_loss,
            "wh_loss": wh_loss,
            "off_loss": off_loss,
            "id_loss": id_loss,
        }
        return loss, loss_stats


class MotTrainer(BaseTrainer):
    def __init__(
        self,
        model,
        gpus,
        num_iters,
        exp_id,
        device,
        hide_data_time,
        print_iter,
        optimizer,
        mse_loss,
        reg_loss,
        dense_wh,
        cat_spec_wh,
        reid_dim,
        nID,
        norm_wh,
        num_stack,
        wh_weight,
        off_weight,
        id_weight,
        reg_offset,
        hm_weight,
    ):
        super(MotTrainer, self).__init__(
            model,
            gpus,
            num_iters,
            exp_id,
            device,
            hide_data_time,
            print_iter,
            optimizer,
            mse_loss,
            reg_loss,
            dense_wh,
            cat_spec_wh,
            reid_dim,
            nID,
            norm_wh,
            num_stack,
            wh_weight,
            off_weight,
            id_weight,
            reg_offset,
            hm_weight,
        )

    def _get_losses(
        self,
        mse_loss,
        reg_loss,
        dense_wh,
        cat_spec_wh,
        reid_dim,
        nID,
        norm_wh,
        num_stack,
        wh_weight,
        off_weight,
        id_weight,
        reg_offset,
        hm_weight,
    ):
        loss_states = ["loss", "hm_loss", "wh_loss", "off_loss", "id_loss"]
        loss = MotLoss(
            mse_loss,
            reg_loss,
            dense_wh,
            cat_spec_wh,
            reid_dim,
            nID,
            norm_wh,
            num_stack,
            wh_weight,
            off_weight,
            id_weight,
            reg_offset,
            hm_weight,
        )
        return loss_states, loss

    def save_result(self, output, batch, results):
        reg = output["reg"] if self.opt.reg_offset else None
        dets = mot_decode(
            output["hm"],
            output["wh"],
            reg=reg,
            cat_spec_wh=self.opt.cat_spec_wh,
            K=self.opt.K,
        )
        dets = dets.detach().cpu().numpy().reshape(1, -1, dets.shape[2])
        dets_out = ctdet_post_process(
            dets.copy(),
            batch["meta"]["c"].cpu().numpy(),
            batch["meta"]["s"].cpu().numpy(),
            output["hm"].shape[2],
            output["hm"].shape[3],
            output["hm"].shape[1],
        )
        results[batch["meta"]["img_id"].cpu().numpy()[0]] = dets_out[0]



import glob
import random
import time
import os
import os.path as osp

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F
from torchvision.ops import nms

# import maskrcnn_benchmark.layers.nms as nms
# Set printoptions
torch.set_printoptions(linewidth=1320, precision=5, profile="long")
np.set_printoptions(
    linewidth=320, formatter={"float_kind": "{:11.5g}".format}
)  # format short g, %precision=5


def mkdir_if_missing(d):
    if not osp.exists(d):
        os.makedirs(d)


def float3(x):  # format floats to 3 decimals
    return float(format(x, ".3f"))


def init_seeds(seed=0):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def load_classes(path):
    """
    Loads class labels at 'path'
    """
    fp = open(path, "r")
    names = fp.read().split("\n")
    return list(filter(None, names))  # filter removes empty strings (such as last line)


def model_info(model):  # Plots a line-by-line description of a PyTorch model
    n_p = sum(x.numel() for x in model.parameters())  # number parameters
    n_g = sum(
        x.numel() for x in model.parameters() if x.requires_grad
    )  # number gradients
    print(
        "\n%5s %50s %9s %12s %20s %12s %12s"
        % ("layer", "name", "gradient", "parameters", "shape", "mu", "sigma")
    )
    for i, (name, p) in enumerate(model.named_parameters()):
        name = name.replace("module_list.", "")
        print(
            "%5g %50s %9s %12g %20s %12.3g %12.3g"
            % (i, name, p.requires_grad, p.numel(), list(p.shape), p.mean(), p.std())
        )
    print("Model Summary: %g layers, %g parameters, %g gradients\n" % (i + 1, n_p, n_g))


def plot_one_box(
    x, img, color=None, label=None, line_thickness=None
):  # Plots one bounding box on image img
    tl = line_thickness or round(0.0004 * max(img.shape[0:2])) + 1  # line thickness
    color = color or [random.randint(0, 255) for _ in range(3)]
    c1, c2 = (int(x[0]), int(x[1])), (int(x[2]), int(x[3]))
    cv2.rectangle(img, c1, c2, color, thickness=tl)
    if label:
        tf = max(tl - 1, 1)  # font thickness
        t_size = cv2.getTextSize(label, 0, fontScale=tl / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0], c1[1] - t_size[1] - 3
        cv2.rectangle(img, c1, c2, color, -1)  # filled
        cv2.putText(
            img,
            label,
            (c1[0], c1[1] - 2),
            0,
            tl / 3,
            [225, 255, 255],
            thickness=tf,
            lineType=cv2.LINE_AA,
        )


def weights_init_normal(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.03)
    elif classname.find("BatchNorm2d") != -1:
        torch.nn.init.normal_(m.weight.data, 1.0, 0.03)
        torch.nn.init.constant_(m.bias.data, 0.0)


def xyxy2xywh(x):
    # Convert bounding box format from [x1, y1, x2, y2] to [x, y, w, h]
    y = torch.zeros(x.shape) if x.dtype is torch.float32 else np.zeros(x.shape)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2
    y[:, 2] = x[:, 2] - x[:, 0]
    y[:, 3] = x[:, 3] - x[:, 1]
    return y


def xywh2xyxy(x):
    # Convert bounding box format from [x, y, w, h] to [x1, y1, x2, y2]
    y = torch.zeros(x.shape) if x.dtype is torch.float32 else np.zeros(x.shape)
    y[:, 0] = x[:, 0] - x[:, 2] / 2
    y[:, 1] = x[:, 1] - x[:, 3] / 2
    y[:, 2] = x[:, 0] + x[:, 2] / 2
    y[:, 3] = x[:, 1] + x[:, 3] / 2
    return y


def scale_coords(img_size, coords, img0_shape):
    # Rescale x1, y1, x2, y2 from 416 to image size
    gain_w = float(img_size[0]) / img0_shape[1]  # gain  = old / new
    gain_h = float(img_size[1]) / img0_shape[0]
    gain = min(gain_w, gain_h)
    pad_x = (img_size[0] - img0_shape[1] * gain) / 2  # width padding
    pad_y = (img_size[1] - img0_shape[0] * gain) / 2  # height padding
    coords[:, [0, 2]] -= pad_x
    coords[:, [1, 3]] -= pad_y
    coords[:, 0:4] /= gain
    coords[:, :4] = torch.clamp(coords[:, :4], min=0)
    return coords


def ap_per_class(tp, conf, pred_cls, target_cls):
    """ Compute the average precision, given the recall and precision curves.
    Method originally from https://github.com/rafaelpadilla/Object-Detection-Metrics.
    # Arguments
        tp:    True positives (list).
        conf:  Objectness value from 0-1 (list).
        pred_cls: Predicted object classes (list).
        target_cls: True object classes (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """

    # lists/pytorch to numpy
    tp, conf, pred_cls, target_cls = (
        np.array(tp),
        np.array(conf),
        np.array(pred_cls),
        np.array(target_cls),
    )

    # Sort by objectness
    i = np.argsort(-conf)
    tp, conf, pred_cls = tp[i], conf[i], pred_cls[i]

    # Find unique classes
    unique_classes = np.unique(np.concatenate((pred_cls, target_cls), 0))

    # Create Precision-Recall curve and compute AP for each class
    ap, p, r = [], [], []
    for c in unique_classes:
        i = pred_cls == c
        n_gt = sum(target_cls == c)  # Number of ground truth objects
        n_p = sum(i)  # Number of predicted objects

        if (n_p == 0) and (n_gt == 0):
            continue
        elif (n_p == 0) or (n_gt == 0):
            ap.append(0)
            r.append(0)
            p.append(0)
        else:
            # Accumulate FPs and TPs
            fpc = np.cumsum(1 - tp[i])
            tpc = np.cumsum(tp[i])

            # Recall
            recall_curve = tpc / (n_gt + 1e-16)
            r.append(tpc[-1] / (n_gt + 1e-16))

            # Precision
            precision_curve = tpc / (tpc + fpc)
            p.append(tpc[-1] / (tpc[-1] + fpc[-1]))

            # AP from recall-precision curve
            ap.append(compute_ap(recall_curve, precision_curve))

    return np.array(ap), unique_classes.astype("int32"), np.array(r), np.array(p)


def compute_ap(recall, precision):
    """ Compute the average precision, given the recall and precision curves.
    Code originally from https://github.com/rbgirshick/py-faster-rcnn.
    # Arguments
        recall:    The recall curve (list).
        precision: The precision curve (list).
    # Returns
        The average precision as computed in py-faster-rcnn.
    """
    # correct AP calculation
    # first append sentinel values at the end

    mrec = np.concatenate(([0.0], recall, [1.0]))
    mpre = np.concatenate(([0.0], precision, [0.0]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # to calculate area under PR curve, look for points
    # where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * prec
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap


def bbox_iou(box1, box2, x1y1x2y2=False):
    """
    Returns the IoU of two bounding boxes
    """
    N, M = len(box1), len(box2)
    if x1y1x2y2:
        # Get the coordinates of bounding boxes
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[:, 0], box1[:, 1], box1[:, 2], box1[:, 3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[:, 0], box2[:, 1], box2[:, 2], box2[:, 3]
    else:
        # Transform from center and width to exact coordinates
        b1_x1, b1_x2 = box1[:, 0] - box1[:, 2] / 2, box1[:, 0] + box1[:, 2] / 2
        b1_y1, b1_y2 = box1[:, 1] - box1[:, 3] / 2, box1[:, 1] + box1[:, 3] / 2
        b2_x1, b2_x2 = box2[:, 0] - box2[:, 2] / 2, box2[:, 0] + box2[:, 2] / 2
        b2_y1, b2_y2 = box2[:, 1] - box2[:, 3] / 2, box2[:, 1] + box2[:, 3] / 2

    # get the coordinates of the intersection rectangle
    inter_rect_x1 = torch.max(b1_x1.unsqueeze(1), b2_x1)
    inter_rect_y1 = torch.max(b1_y1.unsqueeze(1), b2_y1)
    inter_rect_x2 = torch.min(b1_x2.unsqueeze(1), b2_x2)
    inter_rect_y2 = torch.min(b1_y2.unsqueeze(1), b2_y2)
    # Intersection area
    inter_area = torch.clamp(inter_rect_x2 - inter_rect_x1, 0) * torch.clamp(
        inter_rect_y2 - inter_rect_y1, 0
    )
    # Union Area
    b1_area = (b1_x2 - b1_x1) * (b1_y2 - b1_y1)
    b1_area = ((b1_x2 - b1_x1) * (b1_y2 - b1_y1)).view(-1, 1).expand(N, M)
    b2_area = ((b2_x2 - b2_x1) * (b2_y2 - b2_y1)).view(1, -1).expand(N, M)

    return inter_area / (b1_area + b2_area - inter_area + 1e-16)


def build_targets_max(target, anchor_wh, nA, nC, nGh, nGw):
    """
    returns nT, nCorrect, tx, ty, tw, th, tconf, tcls
    """
    nB = len(target)  # number of images in batch

    txy = torch.zeros(nB, nA, nGh, nGw, 2).cuda()  # batch size, anchors, grid size
    twh = torch.zeros(nB, nA, nGh, nGw, 2).cuda()
    tconf = torch.LongTensor(nB, nA, nGh, nGw).fill_(0).cuda()
    tcls = (
        torch.ByteTensor(nB, nA, nGh, nGw, nC).fill_(0).cuda()
    )  # nC = number of classes
    tid = torch.LongTensor(nB, nA, nGh, nGw, 1).fill_(-1).cuda()
    for b in range(nB):
        t = target[b]
        t_id = t[:, 1].clone().long().cuda()
        t = t[:, [0, 2, 3, 4, 5]]
        nTb = len(t)  # number of targets
        if nTb == 0:
            continue

        gxy, gwh = t[:, 1:3].clone(), t[:, 3:5].clone()
        gxy[:, 0] = gxy[:, 0] * nGw
        gxy[:, 1] = gxy[:, 1] * nGh
        gwh[:, 0] = gwh[:, 0] * nGw
        gwh[:, 1] = gwh[:, 1] * nGh
        gi = torch.clamp(gxy[:, 0], min=0, max=nGw - 1).long()
        gj = torch.clamp(gxy[:, 1], min=0, max=nGh - 1).long()

        # iou of targets-anchors (using wh only)
        box1 = gwh
        box2 = anchor_wh.unsqueeze(1)
        inter_area = torch.min(box1, box2).prod(2)
        iou = inter_area / (box1.prod(1) + box2.prod(2) - inter_area + 1e-16)

        # Select best iou_pred and anchor
        iou_best, a = iou.max(0)  # best anchor [0-2] for each target

        # Select best unique target-anchor combinations
        if nTb > 1:
            _, iou_order = torch.sort(-iou_best)  # best to worst

            # Unique anchor selection
            u = torch.stack((gi, gj, a), 0)[:, iou_order]
            first_unique = return_torch_unique_index(
                u, torch.unique(u, dim=1)
            )  # torch alternative
            i = iou_order[first_unique]
            # best anchor must share significant commonality (iou) with target
            i = i[iou_best[i] > 0.60]
            if len(i) == 0:
                continue

            a, gj, gi, t = a[i], gj[i], gi[i], t[i]
            t_id = t_id[i]
            if len(t.shape) == 1:
                t = t.view(1, 5)
        else:
            if iou_best < 0.60:
                continue

        tc, gxy, gwh = t[:, 0].long(), t[:, 1:3].clone(), t[:, 3:5].clone()
        gxy[:, 0] = gxy[:, 0] * nGw
        gxy[:, 1] = gxy[:, 1] * nGh
        gwh[:, 0] = gwh[:, 0] * nGw
        gwh[:, 1] = gwh[:, 1] * nGh

        # XY coordinates
        txy[b, a, gj, gi] = gxy - gxy.floor()

        # Width and height
        twh[b, a, gj, gi] = torch.log(gwh / anchor_wh[a])  # yolo method

        # One-hot encoding of label
        tcls[b, a, gj, gi, tc] = 1
        tconf[b, a, gj, gi] = 1
        tid[b, a, gj, gi] = t_id.unsqueeze(1)
    tbox = torch.cat([txy, twh], -1)
    return tconf, tbox, tid


def generate_anchor(nGh, nGw, anchor_wh):
    nA = len(anchor_wh)
    yy, xx = torch.meshgrid(torch.arange(nGh), torch.arange(nGw))
    xx, yy = xx.cuda(), yy.cuda()

    mesh = torch.stack([xx, yy], dim=0)  # Shape 2, nGh, nGw
    mesh = mesh.unsqueeze(0).repeat(nA, 1, 1, 1).float()  # Shape nA x 2 x nGh x nGw
    anchor_offset_mesh = (
        anchor_wh.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, nGh, nGw)
    )  # Shape nA x 2 x nGh x nGw
    anchor_mesh = torch.cat(
        [mesh, anchor_offset_mesh], dim=1
    )  # Shape nA x 4 x nGh x nGw
    return anchor_mesh


def encode_delta(gt_box_list, fg_anchor_list):
    px, py, pw, ph = (
        fg_anchor_list[:, 0],
        fg_anchor_list[:, 1],
        fg_anchor_list[:, 2],
        fg_anchor_list[:, 3],
    )
    gx, gy, gw, gh = (
        gt_box_list[:, 0],
        gt_box_list[:, 1],
        gt_box_list[:, 2],
        gt_box_list[:, 3],
    )
    dx = (gx - px) / pw
    dy = (gy - py) / ph
    dw = torch.log(gw / pw)
    dh = torch.log(gh / ph)
    return torch.stack([dx, dy, dw, dh], dim=1)


def decode_delta(delta, fg_anchor_list):
    px, py, pw, ph = (
        fg_anchor_list[:, 0],
        fg_anchor_list[:, 1],
        fg_anchor_list[:, 2],
        fg_anchor_list[:, 3],
    )
    dx, dy, dw, dh = delta[:, 0], delta[:, 1], delta[:, 2], delta[:, 3]
    gx = pw * dx + px
    gy = ph * dy + py
    gw = pw * torch.exp(dw)
    gh = ph * torch.exp(dh)
    return torch.stack([gx, gy, gw, gh], dim=1)


def decode_delta_map(delta_map, anchors):
    """
    :param: delta_map, shape (nB, nA, nGh, nGw, 4)
    :param: anchors, shape (nA,4)
    """
    nB, nA, nGh, nGw, _ = delta_map.shape
    anchor_mesh = generate_anchor(nGh, nGw, anchors)
    anchor_mesh = anchor_mesh.permute(
        0, 2, 3, 1
    ).contiguous()  # Shpae (nA x nGh x nGw) x 4
    anchor_mesh = anchor_mesh.unsqueeze(0).repeat(nB, 1, 1, 1, 1)
    pred_list = decode_delta(delta_map.view(-1, 4), anchor_mesh.view(-1, 4))
    pred_map = pred_list.view(nB, nA, nGh, nGw, 4)
    return pred_map


def pooling_nms(heatmap, kernel=1):
    pad = (kernel - 1) // 2
    hmax = F.max_pool2d(heatmap, (kernel, kernel), stride=1, padding=pad)
    keep = (hmax == heatmap).float()
    return keep * heatmap


def non_max_suppression(prediction, conf_thres=0.5, nms_thres=0.2):
    """
    Removes detections with lower object confidence score than 'conf_thres'
    Non-Maximum Suppression to further filter detections.
    Returns detections with shape:
        (x1, y1, x2, y2, object_conf, class_score, class_pred)
    """

    output = [None for _ in range(len(prediction))]
    for image_i, pred in enumerate(prediction):
        # Filter out confidence scores below threshold
        # Get score and class with highest confidence

        v = pred[:, 4] > conf_thres
        v = v.nonzero().squeeze()
        if len(v.shape) == 0:
            v = v.unsqueeze(0)

        pred = pred[v]

        # If none are remaining => process next image
        nP = pred.shape[0]
        if not nP:
            continue
        # From (center x, center y, width, height) to (x1, y1, x2, y2)
        pred[:, :4] = xywh2xyxy(pred[:, :4])
        nms_indices = nms(pred[:, :4], pred[:, 4], nms_thres)
        det_max = pred[nms_indices]

        if len(det_max) > 0:
            # Add max detections to outputs
            output[image_i] = (
                det_max
                if output[image_i] is None
                else torch.cat((output[image_i], det_max))
            )

    return output


def return_torch_unique_index(u, uv):
    n = uv.shape[1]  # number of columns
    first_unique = torch.zeros(n, device=u.device).long()
    for j in range(n):
        first_unique[j] = (uv[:, j : j + 1] == u).all(0).nonzero()[0]

    return first_unique


def strip_optimizer_from_checkpoint(filename="weights/best.pt"):
    # Strip optimizer from *.pt files for lighter files (reduced by 2/3 size)

    a = torch.load(filename, map_location="cpu")
    a["optimizer"] = []
    torch.save(a, filename.replace(".pt", "_lite.pt"))


def plot_results():
    # Plot YOLO training results file 'results.txt'
    # import os; os.system('wget https://storage.googleapis.com/ultralytics/yolov3/results_v1.txt')

    plt.figure(figsize=(14, 7))
    s = [
        "X + Y",
        "Width + Height",
        "Confidence",
        "Classification",
        "Total Loss",
        "mAP",
        "Recall",
        "Precision",
    ]
    files = sorted(glob.glob("results*.txt"))
    for f in files:
        results = np.loadtxt(
            f, usecols=[2, 3, 4, 5, 6, 9, 10, 11]
        ).T  # column 11 is mAP
        x = range(1, results.shape[1])
        for i in range(8):
            plt.subplot(2, 4, i + 1)
            plt.plot(x, results[i, x], marker=".", label=f)
            plt.title(s[i])
            if i == 0:
                plt.legend()



import os
from typing import Dict
import numpy as np

from tracking_utils.log import logger


def write_results(filename, results_dict: Dict, data_type: str):
    if not filename:
        return
    path = os.path.dirname(filename)
    if not os.path.exists(path):
        os.makedirs(path)

    if data_type in ("mot", "mcmot", "lab"):
        save_format = "{frame},{id},{x1},{y1},{w},{h},1,-1,-1,-1\n"
    elif data_type == "kitti":
        save_format = "{frame} {id} pedestrian -1 -1 -10 {x1} {y1} {x2} {y2} -1 -1 -1 -1000 -1000 -1000 -10 {score}\n"
    else:
        raise ValueError(data_type)

    with open(filename, "w") as f:
        for frame_id, frame_data in results_dict.items():
            if data_type == "kitti":
                frame_id -= 1
            for tlwh, track_id in frame_data:
                if track_id < 0:
                    continue
                x1, y1, w, h = tlwh
                x2, y2 = x1 + w, y1 + h
                line = save_format.format(
                    frame=frame_id,
                    id=track_id,
                    x1=x1,
                    y1=y1,
                    x2=x2,
                    y2=y2,
                    w=w,
                    h=h,
                    score=1.0,
                )
                f.write(line)
    logger.info("Save results to {}".format(filename))


def read_results(filename, data_type: str, is_gt=False, is_ignore=False):
    if data_type in ("mot", "lab"):
        read_fun = read_mot_results
    else:
        raise ValueError("Unknown data type: {}".format(data_type))

    return read_fun(filename, is_gt, is_ignore)


"""
labels={'ped', ...			% 1
'person_on_vhcl', ...	% 2
'car', ...				% 3
'bicycle', ...			% 4
'mbike', ...			% 5
'non_mot_vhcl', ...		% 6
'static_person', ...	% 7
'distractor', ...		% 8
'occluder', ...			% 9
'occluder_on_grnd', ...		%10
'occluder_full', ...		% 11
'reflection', ...		% 12
'crowd' ...			% 13
};
"""


def read_mot_results(filename, is_gt, is_ignore):
    valid_labels = {1}
    ignore_labels = {2, 7, 8, 12}
    results_dict = dict()
    if os.path.isfile(filename):
        with open(filename, "r") as f:
            for line in f.readlines():
                linelist = line.split(",")
                if len(linelist) < 7:
                    continue
                fid = int(linelist[0])
                if fid < 1:
                    continue
                results_dict.setdefault(fid, list())

                box_size = float(linelist[4]) * float(linelist[5])

                if is_gt:
                    if "MOT16-" in filename or "MOT17-" in filename:
                        label = int(float(linelist[7]))
                        mark = int(float(linelist[6]))
                        if mark == 0 or label not in valid_labels:
                            continue
                    score = 1
                elif is_ignore:
                    if "MOT16-" in filename or "MOT17-" in filename:
                        label = int(float(linelist[7]))
                        vis_ratio = float(linelist[8])
                        if label not in ignore_labels and vis_ratio >= 0:
                            continue
                    else:
                        continue
                    score = 1
                else:
                    score = float(linelist[6])

                tlwh = tuple(map(float, linelist[2:6]))
                target_id = int(linelist[1])

                results_dict[fid].append((tlwh, target_id, score))

    return results_dict


def unzip_objs(objs):
    if len(objs) > 0:
        tlwhs, ids, scores = zip(*objs)
    else:
        tlwhs, ids, scores = [], [], []
    tlwhs = np.asarray(tlwhs, dtype=float).reshape(-1, 4)

    return tlwhs, ids, scores



import numpy as np
import torch

from .deep.feature_extractor import Extractor
from .sort.nn_matching import NearestNeighborDistanceMetric
from .sort.preprocessing import non_max_suppression
from .sort.detection import Detection
from .sort.tracker import Tracker


__all__ = ["DeepSort"]


class DeepSort(object):
    def __init__(
        self,
        max_dist=0.2,
        min_confidence=0.3,
        nms_max_overlap=1.0,
        max_iou_distance=0.7,
        max_age=70,
        n_init=3,
        nn_budget=100,
        device=None,
    ):
        self.min_confidence = min_confidence
        self.nms_max_overlap = nms_max_overlap

        self.extractor = Extractor(device=device)

        max_cosine_distance = max_dist
        nn_budget = 100
        metric = NearestNeighborDistanceMetric("cosine", max_cosine_distance, nn_budget)
        self.tracker = Tracker(
            metric, max_iou_distance=max_iou_distance, max_age=max_age, n_init=n_init
        )

    def update(self, bbox_xywh, confidences, cls_ids, ori_img, is_new):
        self.width, self.height = ori_img.shape[:2]
        # generate detections
        features = self._get_features(bbox_xywh, ori_img)
        bbox_tlwh = self._xywh_to_tlwh(bbox_xywh)
        detections = [
            Detection(bbox_tlwh[i], conf, cls_ids[i], features[i])
            for i, conf in enumerate(confidences)
            if conf > self.min_confidence
        ]

        # run on non-maximum supression
        boxes = np.array([d.tlwh for d in detections])
        scores = np.array([d.confidence for d in detections])
        indices = non_max_suppression(boxes, self.nms_max_overlap, scores)
        detections = [detections[i] for i in indices]

        # update tracker
        self.tracker.predict()
        self.tracker.update(detections)

        # output bbox identities
        outputs = []
        for track in self.tracker.tracks:
            if not is_new and (not track.is_confirmed() or track.time_since_update > 1):
                continue
            box = track.to_tlwh()
            cls_id = track.cls_id
            confidence = track.confidence
            x1, y1, x2, y2 = self._tlwh_to_xyxy(box)
            track_id = track.track_id
            outputs.append(np.array([x1, y1, x2, y2, track_id, confidence, cls_id], dtype=np.int))
        if len(outputs) > 0:
            outputs = np.stack(outputs, axis=0)
        return outputs

    """
    TODO:
        Convert bbox from xc_yc_w_h to xtl_ytl_w_h
    Thanks JieChen91@github.com for reporting this bug!
    """

    @staticmethod
    def _xywh_to_tlwh(bbox_xywh):
        if isinstance(bbox_xywh, np.ndarray):
            bbox_tlwh = bbox_xywh.copy()
        elif isinstance(bbox_xywh, torch.Tensor):
            bbox_tlwh = bbox_xywh.clone()
        bbox_tlwh[:, 0] = bbox_xywh[:, 0] - bbox_xywh[:, 2] / 2.0
        bbox_tlwh[:, 1] = bbox_xywh[:, 1] - bbox_xywh[:, 3] / 2.0
        return bbox_tlwh

    def _xywh_to_xyxy(self, bbox_xywh):
        x, y, w, h = bbox_xywh
        x1 = max(int(x - w / 2), 0)
        x2 = min(int(x + w / 2), self.width - 1)
        y1 = max(int(y - h / 2), 0)
        y2 = min(int(y + h / 2), self.height - 1)
        return x1, y1, x2, y2

    def _tlwh_to_xyxy(self, bbox_tlwh):
        """
        TODO:
            Convert bbox from xtl_ytl_w_h to xc_yc_w_h
        Thanks JieChen91@github.com for reporting this bug!
        """
        x, y, w, h = bbox_tlwh
        x1 = max(int(x), 0)
        x2 = min(int(x + w), self.width - 1)
        y1 = max(int(y), 0)
        y2 = min(int(y + h), self.height - 1)
        return x1, y1, x2, y2

    def _xyxy_to_tlwh(self, bbox_xyxy):
        x1, y1, x2, y2 = bbox_xyxy

        t = x1
        l = y1
        w = int(x2 - x1)
        h = int(y2 - y1)
        return t, l, w, h

    def _get_features(self, bbox_xywh, ori_img):
        im_crops = []
        for box in bbox_xywh:
            x1, y1, x2, y2 = self._tlwh_to_xyxy(box)
            im = ori_img[x1:x2, y1:y2]
            im_crops.append(im)
        if im_crops:
            features = self.extractor(im_crops)
        else:
            features = np.array([])
        return features



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch
import torch.nn as nn
from .utils import _gather_feat, _tranpose_and_gather_feat


def _nms(heat, kernel=3):
    pad = (kernel - 1) // 2

    hmax = nn.functional.max_pool2d(heat, (kernel, kernel), stride=1, padding=pad)
    keep = (hmax == heat).float()
    return heat * keep


def _topk_channel(scores, K=40):
    batch, cat, height, width = scores.size()

    topk_scores, topk_inds = torch.topk(scores.view(batch, cat, -1), K)

    topk_inds = topk_inds % (height * width)
    topk_ys = (topk_inds / width).int().float()
    topk_xs = (topk_inds % width).int().float()

    return topk_scores, topk_inds, topk_ys, topk_xs


def _topk(scores, K=40):
    batch, cat, height, width = scores.size()

    topk_scores, topk_inds = torch.topk(scores.view(batch, cat, -1), K)

    topk_inds = topk_inds % (height * width)
    topk_ys = (topk_inds / width).int().float()
    topk_xs = (topk_inds % width).int().float()

    topk_score, topk_ind = torch.topk(topk_scores.view(batch, -1), K)
    topk_clses = (topk_ind / K).int()
    topk_inds = _gather_feat(topk_inds.view(batch, -1, 1), topk_ind).view(batch, K)
    topk_ys = _gather_feat(topk_ys.view(batch, -1, 1), topk_ind).view(batch, K)
    topk_xs = _gather_feat(topk_xs.view(batch, -1, 1), topk_ind).view(batch, K)

    return topk_score, topk_inds, topk_clses, topk_ys, topk_xs


def mot_decode(heat, wh, reg=None, ltrb=False, K=100):
    batch, cat, height, width = heat.size()

    # perform nms on heatmaps
    heat = _nms(heat)

    scores, inds, clses, ys, xs = _topk(heat, K=K)
    if reg is not None:
        reg = _tranpose_and_gather_feat(reg, inds)
        reg = reg.view(batch, K, 2)
        xs = xs.view(batch, K, 1) + reg[:, :, 0:1]
        ys = ys.view(batch, K, 1) + reg[:, :, 1:2]
    else:
        xs = xs.view(batch, K, 1) + 0.5
        ys = ys.view(batch, K, 1) + 0.5
    wh = _tranpose_and_gather_feat(wh, inds)
    if ltrb:
        wh = wh.view(batch, K, 4)
    else:
        wh = wh.view(batch, K, 2)
    clses = clses.view(batch, K, 1).float()
    scores = scores.view(batch, K, 1)
    if ltrb:
        bboxes = torch.cat(
            [
                xs - wh[..., 0:1],
                ys - wh[..., 1:2],
                xs + wh[..., 2:3],
                ys + wh[..., 3:4],
            ],
            dim=2,
        )
    else:
        bboxes = torch.cat(
            [
                xs - wh[..., 0:1] / 2,
                ys - wh[..., 1:2] / 2,
                xs + wh[..., 0:1] / 2,
                ys + wh[..., 1:2] / 2,
            ],
            dim=2,
        )
    detections = torch.cat([bboxes, scores, clses], dim=2)

    return detections, inds



import os.path as osp
import os
import numpy as np


def gen_labels_mot(
    seq_root,
    label_root
):

    def mkdirs(d):
        if not osp.exists(d):
            os.makedirs(d)

    mkdirs(label_root)
    seqs = [s for s in os.listdir(seq_root)]

    tid_curr = 0
    tid_last = -1
    for seq in seqs:
        with open(osp.join(seq_root, seq, "seqinfo.ini")) as f:
            seq_info = f.read()
        seq_width = int(
            seq_info[seq_info.find("imWidth=") + 8: seq_info.find("\nimHeight")]
        )
        seq_height = int(
            seq_info[seq_info.find("imHeight=") + 9: seq_info.find("\nimExt")]
        )

        gt_txt = osp.join(seq_root, seq, "gt", "gt.txt")
        gt = np.loadtxt(gt_txt, dtype=np.float64, delimiter=",")

        seq_label_root = osp.join(label_root, seq, "img1")
        mkdirs(seq_label_root)

        for fid, tid, x, y, w, h, mark, label, _ in gt:
            if mark == 0 or not label == 1:
                continue
            fid = int(fid)
            tid = int(tid)
            if not tid == tid_last:
                tid_curr += 1
                tid_last = tid
            x += w / 2
            y += h / 2
            label_fpath = osp.join(seq_label_root, "{:06d}.txt".format(fid))
            label_str = "0 {:d} {:.6f} {:.6f} {:.6f} {:.6f}\n".format(
                tid_curr, x / seq_width, y / seq_height, w / seq_width, h / seq_height
            )
            with open(label_fpath, "a") as f:
                f.write(label_str)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import numpy as np
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.trains.mot import MotTrainer
from opendr.perception.object_tracking_2d.logger import Logger
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.tracking_utils.evaluation import Evaluator


def train(
    model,
    infer,
    optimizer,
    train_dataset,
    val_dataset,
    batch_size,
    num_workers,
    gpus,
    chunk_sizes,
    num_iters,
    exp_id,
    device,
    hide_data_time,
    print_iter,
    mse_loss,
    reg_loss,
    dense_wh,
    cat_spec_wh,
    reid_dim,
    nID,
    norm_wh,
    num_stack,
    wh_weight,
    off_weight,
    id_weight,
    num_epochs,
    lr_step,
    save_dir,
    lr,
    reg_offset,
    hm_weight,
    checkpoints_path,
    checkpoint_after_iter,
    start_epoch,
    val_epochs=5,
    log=print,
):

    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=True,
        drop_last=True,
    )

    trainer = MotTrainer(
        model,
        gpus,
        num_iters,
        exp_id,
        device,
        hide_data_time,
        print_iter,
        optimizer,
        mse_loss,
        reg_loss,
        dense_wh,
        cat_spec_wh,
        reid_dim,
        nID,
        norm_wh,
        num_stack,
        wh_weight,
        off_weight,
        id_weight,
        reg_offset,
        hm_weight,
    )
    trainer.set_device(gpus, chunk_sizes, device)

    def save(iter):
        if isinstance(model, torch.nn.DataParallel):
            state_dict = model.module.state_dict()
        else:
            state_dict = model.state_dict()
        data = {'epoch': epoch,
                'state_dict': state_dict}
        if not (optimizer is None):
            data['optimizer'] = optimizer.state_dict()
        torch.save(data, os.path.join(
            checkpoints_path,
            f"checkpoint_{iter}.pth"
        ))

    last_eval_result = {}

    for epoch in range(start_epoch + 1, num_epochs + 1):
        log_dict_train, _ = trainer.train(epoch, train_loader, checkpoint_after_iter, save, log)
        log(Logger.LOG_WHEN_NORMAL, "epoch: {} |".format(epoch))

        if val_epochs > 0:
            if epoch % val_epochs == 0:
                last_eval_result = evaluate(infer, val_dataset, "evaluation " + str(epoch))
                log(Logger.LOG_WHEN_NORMAL, last_eval_result)
        if epoch in lr_step:
            lr = lr * (0.1 ** (lr_step.index(epoch) + 1))
            log(Logger.LOG_WHEN_NORMAL, "Drop LR to", lr)
            for param_group in optimizer.param_groups:
                param_group["lr"] = lr

    print()
    return last_eval_result


def evaluate(
    infer,
    val_dataset,
    name="default",
):

    gt_targets = []
    output_targets = []
    evaluator = Evaluator()

    frame_id = 0

    for frame_id in range(len(val_dataset)):
        image, target = val_dataset[frame_id]
        output_target = infer(image, frame_id)
        output_targets.append(output_target)
        gt_targets.append(target)

    gt_tlwhs = [
        [
            np.array([t.top, t.left, t.width, t.height])
            for t in ts.data
        ] for ts in gt_targets
    ]
    output_tlwhs = [
        [
            np.array([t.top, t.left, t.width, t.height])
            for t in ts.data
        ] for ts in output_targets
    ]

    gt_ids = [
        [
            t.id for t in ts.data
        ] for ts in gt_targets
    ]
    output_ids = [
        [
            t.id for t in ts.data
        ] for ts in output_targets
    ]

    acc = evaluator.eval_all(
        gt_tlwhs,
        gt_ids,
        output_tlwhs,
        output_ids,
    )

    summary = Evaluator.get_summary([acc], [name])

    return summary



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import cv2
import time
import numpy as np
from zipfile import ZipFile
from urllib.request import urlretrieve
from opendr.engine.constants import OPENDR_SERVER_URL
from opendr.engine.data import Image
from opendr.engine.target import Category
from opendr.engine.datasets import ExternalDataset, DatasetIterator


class Market1501Dataset(ExternalDataset):
    def __init__(
        self,
        path,
    ):

        super().__init__(path, "market1501")

        self.path = path

    @staticmethod
    def download(
        url, download_path, dataset_sub_path=".", file_format="zip",
        create_dir=False,
    ):

        if file_format == "zip":
            if create_dir:
                os.makedirs(download_path, exist_ok=True)

            print("Downloading Market1501 Dataset zip file from", url, "to", download_path)

            start_time = 0
            last_print = 0

            def reporthook(count, block_size, total_size):
                nonlocal start_time
                nonlocal last_print
                if count == 0:
                    start_time = time.time()
                    last_print = start_time
                    return

                duration = time.time() - start_time
                progress_size = int(count * block_size)
                speed = int(progress_size / (1024 * duration))
                if time.time() - last_print >= 1:
                    last_print = time.time()
                    print(
                        "\r%d MB, %d KB/s, %d seconds passed" %
                        (progress_size / (1024 * 1024), speed, duration),
                        end=''
                    )

            zip_path = os.path.join(download_path, "dataset.zip")
            urlretrieve(url, zip_path, reporthook=reporthook)
            print()

            print("Extracting Market1501 Dataset from zip file")
            with ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(download_path)

            os.remove(zip_path)

            return Market1501Dataset(
                os.path.join(download_path, dataset_sub_path),
            )
        else:
            raise ValueError("Unsupported file_format: " + file_format)

    @staticmethod
    def download_nano_market1501(
        download_path, create_dir=False,
    ):
        return Market1501Dataset.download(
            os.path.join(OPENDR_SERVER_URL, "perception", "object_tracking_2d", "nano_market1501.zip"),
            download_path,
            dataset_sub_path="nano_market1501",
            create_dir=create_dir,
        )


class Market1501DatasetIterator(DatasetIterator):

    def __init__(
        self,
        path,
    ):
        self.path = path
        self.files = os.listdir(self.path)

    def __getitem__(self, files_index):

        img, label = self.get_data(
            os.path.join(self.path, self.files[files_index]), self.files[files_index]
        )

        return (
            Image(img), Category(label)
        )

    def get_data(self, img_path, label_path):
        img = cv2.imread(img_path)  # BGR
        if img is None:
            raise ValueError("File corrupt {}".format(img_path))

        label = int(label_path.split("_")[0])

        img = np.ascontiguousarray(img[:, :, ::-1])  # BGR to RGB

        return img, label

    def __len__(self):
        return len(self.files)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import time
import torch
import ntpath
import shutil
import numpy as np
import onnxruntime as ort
from torchvision.transforms import transforms as T
from opendr.engine.learners import Learner
from opendr.engine.datasets import DatasetIterator, ExternalDataset, MappedDatasetIterator
from opendr.perception.object_tracking_2d.logger import Logger
from opendr.perception.object_tracking_2d.datasets.mot_dataset import JointDataset, RawMotDatasetIterator
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.models.model import create_model
from opendr.perception.object_tracking_2d.fair_mot.algorithm.run import train, evaluate
from opendr.perception.object_tracking_2d.fair_mot.algorithm.load import load_from_checkpoint
from opendr.perception.object_tracking_2d.datasets.mot_dataset import letterbox, process as process_dataset
from opendr.perception.object_tracking_2d.fair_mot.algorithm.lib.tracker.multitracker import JDETracker
from opendr.engine.data import Image
from opendr.engine.target import TrackingAnnotation, TrackingAnnotationList
from opendr.engine.constants import OPENDR_SERVER_URL
from urllib.request import urlretrieve


class ObjectTracking2DFairMotLearner(Learner):
    def __init__(
        self,
        lr=0.0001,
        iters=-1,
        batch_size=4,
        optimizer="adam",
        lr_schedule="",
        backbone="dla_34",
        network_head="",
        checkpoint_after_iter=0,
        checkpoint_load_iter=0,
        temp_path="",
        device="cuda",
        threshold=0.3,
        scale=1.0,
        lr_step=[20],
        head_conv=256,
        ltrb=True,
        num_classes=1,
        reg_offset=True,
        gpus=[0],
        num_workers=4,
        mse_loss=False,
        reg_loss='l1',
        dense_wh=False,
        cat_spec_wh=False,
        reid_dim=128,
        norm_wh=False,
        wh_weight=0.1,
        off_weight=1,
        id_weight=1,
        num_epochs=30,
        hm_weight=1,
        down_ratio=4,
        max_objs=500,
        track_buffer=30,
        image_mean=[0.408, 0.447, 0.47],
        image_std=[0.289, 0.274, 0.278],
        frame_rate=30,
        min_box_area=100,
        use_pretrained_backbone=True,
    ):
        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        super(ObjectTracking2DFairMotLearner, self).__init__(
            lr=lr,
            iters=iters,
            batch_size=batch_size,
            optimizer=optimizer,
            lr_schedule=lr_schedule,
            backbone=backbone,
            network_head=network_head,
            checkpoint_after_iter=checkpoint_after_iter,
            checkpoint_load_iter=checkpoint_load_iter,
            temp_path=temp_path,
            device=device,
            threshold=threshold,
            scale=scale,
        )

        self.ltrb = ltrb
        self.head_conv = head_conv
        self.num_classes = num_classes
        self.reid_dim = reid_dim
        self.reg_offset = reg_offset
        self.gpus = gpus
        self.num_workers = num_workers
        self.mse_loss = mse_loss
        self.reg_loss = reg_loss
        self.dense_wh = dense_wh
        self.cat_spec_wh = cat_spec_wh
        self.reid_dim = reid_dim
        self.norm_wh = norm_wh
        self.wh_weight = wh_weight
        self.off_weight = off_weight
        self.id_weight = id_weight
        self.num_epochs = num_epochs
        self.lr_step = lr_step
        self.hm_weight = hm_weight
        self.down_ratio = down_ratio
        self.max_objs = max_objs
        self.track_buffer = track_buffer
        self.image_mean = image_mean
        self.image_mean = image_mean
        self.image_std = image_std
        self.frame_rate = frame_rate
        self.min_box_area = min_box_area
        self.use_pretrained_backbone = use_pretrained_backbone

        main_batch_size = self.batch_size // len(self.gpus)
        rest_batch_size = (self.batch_size - main_batch_size)
        self.chunk_sizes = [main_batch_size]
        self.infers_count = 0
        self.infers_time = 0

        for i in range(len(self.gpus) - 1):
            worker_chunk_size = rest_batch_size // (len(self.gpus) - 1)
            if i < rest_batch_size % (len(self.gpus) - 1):
                worker_chunk_size += 1
            self.chunk_sizes.append(worker_chunk_size)

        self.__create_model()

    def save(self, path, verbose=False):
        """
        This method is used to save a trained model.
        Provided with the path, absolute or relative, including a *folder* name, it creates a directory with the name
        of the *folder* provided and saves the model inside with a proper format and a .json file with metadata.
        If self.optimize was ran previously, it saves the optimized ONNX model in a similar fashion, by copying it
        from the self.temp_path it was saved previously during conversion.
        :param path: for the model to be saved, including the folder name
        :type path: str
        :param verbose: whether to print success message or not, defaults to 'False'
        :type verbose: bool, optional
        """

        if self.model is None and self.ort_session is None:
            raise UserWarning("No model is loaded, cannot save.")

        folder_name, _, tail = self.__extract_trailing(path)  # Extract trailing folder name from path
        # Also extract folder name without any extension if extension is erroneously provided
        folder_name_no_ext = folder_name.split(sep='.')[0]

        # Extract path without folder name, by removing folder name from original path
        path_no_folder_name = ''.join(path.rsplit(folder_name, 1))
        # If tail is '', then path was a/b/c/, which leaves a trailing double '/'
        if tail == '':
            path_no_folder_name = path_no_folder_name[0:-1]  # Remove one '/'

        # Create model directory
        new_path = path_no_folder_name + folder_name_no_ext
        os.makedirs(new_path, exist_ok=True)

        model_metadata = {"model_paths": [], "framework": "pytorch", "format": "", "has_data": False,
                          "inference_params": {}, "optimized": None, "optimizer_info": {}}

        if self.model.ort_session is None:
            model_metadata["model_paths"] = [
                folder_name_no_ext + ".pth",
            ]
            model_metadata["optimized"] = False
            model_metadata["format"] = "pth"

            torch.save({
                'state_dict': self.model.state_dict()
            }, os.path.join(path_no_folder_name, folder_name_no_ext, model_metadata["model_paths"][0]))
            if verbose:
                print("Saved Pytorch model.")
        else:
            model_metadata["model_paths"] = [
                folder_name_no_ext + ".onnx"
            ]
            model_metadata["optimized"] = True
            model_metadata["format"] = "onnx"

            shutil.copy2(
                os.path.join(self.temp_path, "onnx_model_temp.onnx"),
                os.path.join(path_no_folder_name, folder_name_no_ext, model_metadata["model_paths"][0])
            )
            if verbose:
                print("Saved ONNX model.")

        with open(os.path.join(new_path, folder_name_no_ext + ".json"), 'w') as outfile:
            json.dump(model_metadata, outfile)

    def load(
        self,
        path,
        verbose=False,
    ):
        """
        Loads the model from inside the path provided, based on the metadata .json file included.
        :param path: path of the directory the model was saved
        :type path: str
        :param verbose: whether to print success message or not, defaults to 'False'
        :type verbose: bool, optional
        """

        model_name, _, _ = self.__extract_trailing(path)  # Trailing folder name from the path provided

        with open(os.path.join(path, model_name + ".json")) as metadata_file:
            metadata = json.load(metadata_file)

        if not metadata["optimized"]:
            self.__load_from_pth(self.model, os.path.join(path, metadata["model_paths"][0]))
            if verbose:
                print("Loaded Pytorch model.")
        else:
            self.__load_rpn_from_onnx(os.path.join(path, metadata["model_paths"][0]))
            if verbose:
                print("Loaded ONNX model.")

    def reset(self):
        self.tracker.reset()

    def fit(
        self,
        dataset,
        val_dataset=None,
        val_epochs=-1,
        logging_path=None,
        silent=False,
        verbose=False,
        train_split_paths=None,
        val_split_paths=None,
        resume_optimizer=False,
        nID=None
    ):

        if train_split_paths is None:
            train_split_paths = {
                "mot20": os.path.join(
                    "perception", "object_tracking_2d", "datasets", "splits", "mot20.train"
                )
            }

        if val_split_paths is None:
            val_split_paths = train_split_paths

        logger = Logger(silent, verbose, logging_path)

        (
            input_dataset_iterator,
            eval_dataset_iterator,
        ) = self._prepare_datasets(
            dataset,
            val_dataset,
            train_split_paths,
            val_split_paths,
            require_val_dataset=val_epochs > 0,
        )

        if nID is None:
            nID = input_dataset_iterator.nID if hasattr(input_dataset_iterator, "nID") else dataset.nID

        checkpoints_path = os.path.join(self.temp_path, "checkpoints")
        if self.checkpoint_after_iter != 0 or self.checkpoint_load_iter != 0:
            os.makedirs(checkpoints_path, exist_ok=True)

        start_epoch = 0

        if self.checkpoint_load_iter != 0:
            _, _, start_epoch = load_from_checkpoint(
                self.model, os.path.join(checkpoints_path, f"checkpoint_{self.checkpoint_load_iter}.pth"),
                self.model_optimizer, resume_optimizer, self.lr, self.lr_step, log=logger.log,
            )

        last_eval_result = train(
            self.model,
            self.infer,
            self.model_optimizer,
            input_dataset_iterator,
            eval_dataset_iterator,
            self.batch_size,
            self.num_workers,
            self.gpus,
            self.chunk_sizes,
            self.iters,
            "train",  # exp_id,
            self.device,
            silent,  # hide_data_time,
            1 if verbose else (-1 if silent else 10),  # print_iter,
            self.mse_loss,
            self.reg_loss,
            self.dense_wh,
            self.cat_spec_wh,
            self.reid_dim,
            nID,
            self.norm_wh,
            1,  # num_stack,
            self.wh_weight,
            self.off_weight,
            self.id_weight,
            self.num_epochs,
            self.lr_step,
            self.temp_path,
            self.lr,
            self.reg_offset,
            self.hm_weight,
            checkpoints_path,
            self.checkpoint_after_iter,
            start_epoch,
            val_epochs=val_epochs,
            log=logger.log,
        )

        logger.close()

        return last_eval_result

    def eval(
        self,
        dataset,
        val_split_paths=None,
        logging_path=None,
        silent=False,
        verbose=False,
    ):

        logger = Logger(silent, verbose, logging_path)

        (
            _,
            eval_dataset_iterator,
        ) = self._prepare_datasets(
            None,
            dataset,
            None,
            val_split_paths,
            require_dataset=False,
        )

        result = evaluate(self.infer, dataset)

        logger.log(Logger.LOG_WHEN_NORMAL, result)

        logger.close()

        return result

    def infer(self, batch, frame_ids=None, img_size=(1088, 608)):

        if self.model is None:
            raise ValueError("No model loaded or created")

        self.model.eval()

        is_single_image = False

        if isinstance(batch, Image):
            batch = [batch]
            is_single_image = True
        elif not isinstance(batch, list):
            raise ValueError("Input batch should be an engine.Image or a list of engine.Image")

        if frame_ids is None:
            frame_ids = [-1] * len(batch)
        elif is_single_image:
            frame_ids = [frame_ids]

        results = []

        for image, frame_id in zip(batch, frame_ids):

            img0 = image.convert("channels_last", "bgr")  # BGR
            img, _, _, _ = letterbox(img0, height=img_size[1], width=img_size[0])

            # Normalize RGB
            img = img[:, :, ::-1].transpose(2, 0, 1)
            img = np.ascontiguousarray(img, dtype=np.float32)
            img /= 255.0

            blob = torch.from_numpy(img).to(self.device).unsqueeze(0)

            t0 = time.time()
            online_targets = self.tracker.update(blob, img0)
            online_tlwhs = []
            online_ids = []
            online_scores = []
            for t in online_targets:
                tlwh = t.tlwh
                tid = t.track_id
                vertical = tlwh[2] / tlwh[3] > 1.6
                if tlwh[2] * tlwh[3] > self.min_box_area and not vertical:
                    online_tlwhs.append(tlwh)
                    online_ids.append(tid)
                    online_scores.append(t.score)

            result = TrackingAnnotationList([
                TrackingAnnotation(
                    name=0,
                    top=tlwh[0],
                    left=tlwh[1],
                    width=tlwh[2],
                    height=tlwh[3],
                    id=id,
                    score=score,
                    frame=frame_id,
                ) for tlwh, id, score in zip(
                    online_tlwhs,
                    online_ids,
                    online_scores
                )
            ])

            t0 = time.time() - t0
            self.infers_count += 1
            self.infers_time += t0

            results.append(result)

        if is_single_image:
            results = results[0]

        return results

    def optimize(self, do_constant_folding=False, img_size=(1088, 608), optimizable_dcn_v2=False):
        """
        Optimize method converts the model to ONNX format and saves the
        model in the parent directory defined by self.temp_path. The ONNX model is then loaded.
        :param do_constant_folding: whether to optimize constants, defaults to 'False'
        :type do_constant_folding: bool, optional
        """

        if not optimizable_dcn_v2:
            raise Exception("Can not optimize the model while DCNv2 implementation is not optimizable")

        if self.model is None:
            raise UserWarning("No model is loaded, cannot optimize. Load or train a model first.")
        if self.model.ort_session is not None:
            raise UserWarning("Model is already optimized in ONNX.")

        input_shape = [
            1,
            3,
            img_size[1],
            img_size[0],
        ]

        try:
            self.__convert_to_onnx(
                input_shape,
                os.path.join(self.temp_path, "onnx_model_temp.onnx"), do_constant_folding
            )
        except FileNotFoundError:
            # Create temp directory
            os.makedirs(self.temp_path, exist_ok=True)
            self.__convert_to_onnx(
                input_shape,
                os.path.join(self.temp_path, "onnx_model_temp.onnx"), do_constant_folding
            )

        self.__load_from_onnx(os.path.join(self.temp_path, "onnx_model_rpn_temp.onnx"))

    @staticmethod
    def download(model_name, path, server_url=None):

        if server_url is None and model_name not in [
            "crowdhuman_dla34",
            "fairmot_dla34",
        ]:
            raise ValueError("Unknown model_name: " + model_name)

        os.makedirs(path, exist_ok=True)

        if server_url is None:
            server_url = os.path.join(
                OPENDR_SERVER_URL, "perception", "object_tracking_2d",
                "fair_mot"
            )

        url = os.path.join(
            server_url, model_name
        )

        model_dir = os.path.join(path, model_name)
        os.makedirs(model_dir, exist_ok=True)

        urlretrieve(os.path.join(
            url, model_name + ".json"
        ), os.path.join(
            model_dir, model_name + ".json"
        ))

        try:
            urlretrieve(os.path.join(
                url, model_name + ".pth"
            ), os.path.join(
                model_dir, model_name + ".pth"
            ))
        except Exception:
            urlretrieve(os.path.join(
                url, model_name + ".tckpt"
            ), os.path.join(
                model_dir, model_name + ".pth"
            ))

        print("Downloaded model", model_name, "to", model_dir)

        return model_dir

    def __convert_to_onnx(self, input_shape, output_name, do_constant_folding=False, verbose=False):
        inp = torch.randn(input_shape).to(self.device)
        input_names = ["data"]
        output_names = self.heads.keys()

        torch.onnx.export(
            self.model, inp, output_name, verbose=verbose, opset_version=11,
            do_constant_folding=do_constant_folding, input_names=input_names, output_names=output_names
        )

    def __load_from_onnx(self, path):
        """
        This method loads an ONNX model from the path provided into an onnxruntime inference session.

        :param path: path to ONNX model
        :type path: str
        """
        self.model.rpn_ort_session = ort.InferenceSession(path)

        # The comments below are the alternative way to use the onnx model, it might be useful in the future
        # depending on how ONNX saving/loading will be implemented across the toolkit.
        # # Load the ONNX model
        # self.model = onnx.load(path)
        #
        # # Check that the IR is well formed
        # onnx.checker.check_model(self.model)
        #
        # # Print a human readable representation of the graph
        # onnx.helper.printable_graph(self.model.graph)

    def __load_from_pth(self, model, path, use_original_dict=False):
        all_params = torch.load(path, map_location=self.device)
        state_dict = all_params['state_dict']
        new_dict = dict()
        for name, tensor in state_dict.items():
            new_name = name.replace('offset_mask', 'offset')
            new_dict[new_name] = tensor
        all_params['state_dict'] = new_dict
        model.load_state_dict(all_params if use_original_dict else all_params["state_dict"])

    def _prepare_datasets(
        self,
        dataset,
        val_dataset,
        train_split_paths,
        val_split_paths,
        require_dataset=True,
        require_val_dataset=True,
    ):

        input_dataset_iterator = None
        eval_dataset_iterator = None

        if isinstance(dataset, ExternalDataset):

            dataset_path = dataset.path
            if dataset.dataset_type.lower() != "mot":
                raise ValueError(
                    "ExternalDataset (" + str(dataset) +
                    ") is given as a dataset, but it is not a MOT dataset")

            transforms = T.Compose([T.ToTensor()])
            input_dataset_iterator = JointDataset(
                dataset_path,
                train_split_paths,
                down_ratio=self.down_ratio,
                max_objects=self.max_objs,
                ltrb=self.ltrb,
                mse_loss=self.mse_loss,
                augment=False, transforms=transforms,
            )
        elif isinstance(dataset, DatasetIterator):
            input_dataset_iterator = MappedDatasetIterator(
                dataset,
                lambda d: process_dataset(
                    d[0], d[1], self.ltrb, self.down_ratio,
                    self.max_objs, self.num_classes, self.mse_loss
                )
            )
        else:
            if require_dataset or dataset is not None:
                raise ValueError(
                    "dataset parameter should be an ExternalDataset or a DatasetIterator"
                )

        if isinstance(val_dataset, ExternalDataset):

            val_dataset_path = val_dataset.path
            if val_dataset.dataset_type.lower() != "mot":
                raise ValueError(
                    "ExternalDataset (" + str(val_dataset) +
                    ") is given as a val_dataset, but it is not a MOT dataset"
                )

            eval_dataset_iterator = RawMotDatasetIterator(
                val_dataset_path,
                val_split_paths,
                down_ratio=self.down_ratio,
                max_objects=self.max_objs,
                ltrb=self.ltrb,
                mse_loss=self.mse_loss,
            )

        elif isinstance(val_dataset, DatasetIterator):
            eval_dataset_iterator = val_dataset
        elif val_dataset is None:
            if isinstance(dataset, ExternalDataset):
                val_dataset_path = dataset.path
                if dataset.dataset_type.lower() != "mot":
                    raise ValueError(
                        "ExternalDataset (" + str(dataset) +
                        ") is given as a dataset, but it is not a MOT dataset"
                    )

                eval_dataset_iterator = RawMotDatasetIterator(
                    val_dataset_path,
                    val_split_paths,
                    down_ratio=self.down_ratio,
                    max_objects=self.max_objs,
                    ltrb=self.ltrb,
                    mse_loss=self.mse_loss,
                )

            elif require_val_dataset:
                raise ValueError(
                    "val_dataset is None and can't be derived from" +
                    " the dataset object because the dataset is not an ExternalDataset"
                )
            else:
                eval_dataset_iterator = input_dataset_iterator
        else:
            raise ValueError(
                "val_dataset parameter should be an ExternalDataset or a DatasetIterator or None"
            )

        return input_dataset_iterator, eval_dataset_iterator

    def __create_model(self):

        heads = {
            'hm': self.num_classes,
            'wh': 2 if not self.ltrb else 4,
            'id': self.reid_dim
        }
        if self.reg_offset:
            heads.update({'reg': 2})

        self.heads = heads

        self.model = create_model(self.use_pretrained_backbone, self.backbone, heads, self.head_conv)
        self.model.to(self.device)
        self.model.ort_session = None
        self.model.heads_names = heads.keys()

        self.model_optimizer = torch.optim.Adam(self.model.parameters(), self.lr)

        self.tracker = JDETracker(
            self.model,
            self.threshold,
            self.track_buffer,
            self.max_objs,
            self.image_mean,
            self.image_std,
            self.down_ratio,
            self.num_classes,
            self.reg_offset,
            self.ltrb,
            self.frame_rate,
        )

    @staticmethod
    def __extract_trailing(path):
        """
        Extracts the trailing folder name or filename from a path provided in an OS-generic way, also handling
        cases where the last trailing character is a separator. Returns the folder name and the split head and tail.
        :param path: the path to extract the trailing filename or folder name from
        :type path: str
        :return: the folder name, the head and tail of the path
        :rtype: tuple of three strings
        """
        head, tail = ntpath.split(path)
        folder_name = tail or ntpath.basename(head)  # handle both a/b/c and a/b/c/
        return folder_name, head, tail



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import logging
import numpy as np
from os.path import join

import torch
from torch import nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo

from DCN import DCN

BN_MOMENTUM = 0.1
logger = logging.getLogger(__name__)


def get_model_url(data="imagenet", name="dla34", hash="ba72cf86"):
    return join("http://dl.yf.io/dla/models", data, "{}-{}.pth".format(name, hash))


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False,
    )


class BasicBlock(nn.Module):
    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            inplanes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(Bottleneck, self).__init__()
        expansion = Bottleneck.expansion
        bottle_planes = planes // expansion
        self.conv1 = nn.Conv2d(inplanes, bottle_planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(bottle_planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(
            bottle_planes,
            bottle_planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn2 = nn.BatchNorm2d(bottle_planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(bottle_planes, planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += residual
        out = self.relu(out)

        return out


class BottleneckX(nn.Module):
    expansion = 2
    cardinality = 32

    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(BottleneckX, self).__init__()
        cardinality = BottleneckX.cardinality
        bottle_planes = planes * cardinality // 32
        self.conv1 = nn.Conv2d(inplanes, bottle_planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(bottle_planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(
            bottle_planes,
            bottle_planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
            groups=cardinality,
        )
        self.bn2 = nn.BatchNorm2d(bottle_planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(bottle_planes, planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += residual
        out = self.relu(out)

        return out


class Root(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, residual):
        super(Root, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            1,
            stride=1,
            bias=False,
            padding=(kernel_size - 1) // 2,
        )
        self.bn = nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.residual = residual

    def forward(self, *x):
        children = x
        x = self.conv(torch.cat(x, 1))
        x = self.bn(x)
        if self.residual:
            x += children[0]
        x = self.relu(x)

        return x


class Tree(nn.Module):
    def __init__(
        self,
        levels,
        block,
        in_channels,
        out_channels,
        stride=1,
        level_root=False,
        root_dim=0,
        root_kernel_size=1,
        dilation=1,
        root_residual=False,
    ):
        super(Tree, self).__init__()
        if root_dim == 0:
            root_dim = 2 * out_channels
        if level_root:
            root_dim += in_channels
        if levels == 1:
            self.tree1 = block(in_channels, out_channels, stride, dilation=dilation)
            self.tree2 = block(out_channels, out_channels, 1, dilation=dilation)
        else:
            self.tree1 = Tree(
                levels - 1,
                block,
                in_channels,
                out_channels,
                stride,
                root_dim=0,
                root_kernel_size=root_kernel_size,
                dilation=dilation,
                root_residual=root_residual,
            )
            self.tree2 = Tree(
                levels - 1,
                block,
                out_channels,
                out_channels,
                root_dim=root_dim + out_channels,
                root_kernel_size=root_kernel_size,
                dilation=dilation,
                root_residual=root_residual,
            )
        if levels == 1:
            self.root = Root(root_dim, out_channels, root_kernel_size, root_residual)
        self.level_root = level_root
        self.root_dim = root_dim
        self.downsample = None
        self.project = None
        self.levels = levels
        if stride > 1:
            self.downsample = nn.MaxPool2d(stride, stride=stride)
        if in_channels != out_channels:
            self.project = nn.Sequential(
                nn.Conv2d(
                    in_channels, out_channels, kernel_size=1, stride=1, bias=False,
                ),
                nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM),
            )

    def forward(self, x, residual=None, children=None):
        children = [] if children is None else children
        bottom = self.downsample(x) if self.downsample else x
        residual = self.project(bottom) if self.project else bottom
        if self.level_root:
            children.append(bottom)
        x1 = self.tree1(x, residual)
        if self.levels == 1:
            x2 = self.tree2(x1)
            x = self.root(x2, x1, *children)
        else:
            children.append(x1)
            x = self.tree2(x1, children=children)
        return x


class DLA(nn.Module):
    def __init__(
        self,
        levels,
        channels,
        num_classes=1000,
        block=BasicBlock,
        residual_root=False,
        linear_root=False,
    ):
        super(DLA, self).__init__()
        self.channels = channels
        self.num_classes = num_classes
        self.base_layer = nn.Sequential(
            nn.Conv2d(3, channels[0], kernel_size=7, stride=1, padding=3, bias=False),
            nn.BatchNorm2d(channels[0], momentum=BN_MOMENTUM),
            nn.ReLU(inplace=True),
        )
        self.level0 = self._make_conv_level(channels[0], channels[0], levels[0])
        self.level1 = self._make_conv_level(
            channels[0], channels[1], levels[1], stride=2
        )
        self.level2 = Tree(
            levels[2],
            block,
            channels[1],
            channels[2],
            2,
            level_root=False,
            root_residual=residual_root,
        )
        self.level3 = Tree(
            levels[3],
            block,
            channels[2],
            channels[3],
            2,
            level_root=True,
            root_residual=residual_root,
        )
        self.level4 = Tree(
            levels[4],
            block,
            channels[3],
            channels[4],
            2,
            level_root=True,
            root_residual=residual_root,
        )
        self.level5 = Tree(
            levels[5],
            block,
            channels[4],
            channels[5],
            2,
            level_root=True,
            root_residual=residual_root,
        )

    def _make_level(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes:
            downsample = nn.Sequential(
                nn.MaxPool2d(stride, stride=stride),
                nn.Conv2d(inplanes, planes, kernel_size=1, stride=1, bias=False),
                nn.BatchNorm2d(planes, momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample=downsample))
        for i in range(1, blocks):
            layers.append(block(inplanes, planes))

        return nn.Sequential(*layers)

    def _make_conv_level(self, inplanes, planes, convs, stride=1, dilation=1):
        modules = []
        for i in range(convs):
            modules.extend(
                [
                    nn.Conv2d(
                        inplanes,
                        planes,
                        kernel_size=3,
                        stride=stride if i == 0 else 1,
                        padding=dilation,
                        bias=False,
                        dilation=dilation,
                    ),
                    nn.BatchNorm2d(planes, momentum=BN_MOMENTUM),
                    nn.ReLU(inplace=True),
                ]
            )
            inplanes = planes
        return nn.Sequential(*modules)

    def forward(self, x):
        y = []
        x = self.base_layer(x)
        for i in range(6):
            x = getattr(self, "level{}".format(i))(x)
            y.append(x)
        return y

    def load_model(self, pretrained=True, data="imagenet", name="dla34", hash="ba72cf86", num_classes=1000):
        
        if pretrained:
            if name.endswith(".pth"):
                model_weights = torch.load(data + name)
            else:
                model_url = get_model_url(data, name, hash)
                model_weights = model_zoo.load_url(model_url)
            num_classes = len(model_weights[list(model_weights.keys())[-1]])

        self.fc = nn.Conv2d(
            self.channels[-1],
            num_classes,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=True,
        )

        if pretrained:
            self.load_state_dict(model_weights)


def dla34(pretrained=True, **kwargs):  # DLA-34
    model = DLA(
        [1, 1, 1, 2, 2, 1], [16, 32, 64, 128, 256, 512], block=BasicBlock, **kwargs
    )
    model.load_model(data="imagenet", name="dla34", hash="ba72cf86", pretrained=pretrained)
    return model


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


def fill_fc_weights(layers):
    for m in layers.modules():
        if isinstance(m, nn.Conv2d):
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)


def fill_up_weights(up):
    w = up.weight.data
    f = math.ceil(w.size(2) / 2)
    c = (2 * f - 1 - f % 2) / (2.0 * f)
    for i in range(w.size(2)):
        for j in range(w.size(3)):
            w[0, 0, i, j] = (1 - math.fabs(i / f - c)) * (1 - math.fabs(j / f - c))
    for c in range(1, w.size(0)):
        w[c, 0, :, :] = w[0, 0, :, :]


class DeformConv(nn.Module):
    def __init__(self, chi, cho):
        super(DeformConv, self).__init__()
        self.actf = nn.Sequential(
            nn.BatchNorm2d(cho, momentum=BN_MOMENTUM), nn.ReLU(inplace=True)
        )
        self.conv = DCN(
            chi,
            cho,
            kernel_size=(3, 3),
            stride=1,
            padding=1,
            dilation=1,
            deformable_groups=1,
        )

    def forward(self, x):
        x = self.conv(x)
        x = self.actf(x)
        return x


class IDAUp(nn.Module):
    def __init__(self, o, channels, up_f):
        super(IDAUp, self).__init__()
        for i in range(1, len(channels)):
            c = channels[i]
            f = int(up_f[i])
            proj = DeformConv(c, o)
            node = DeformConv(o, o)

            up = nn.ConvTranspose2d(
                o,
                o,
                f * 2,
                stride=f,
                padding=f // 2,
                output_padding=0,
                groups=o,
                bias=False,
            )
            fill_up_weights(up)

            setattr(self, "proj_" + str(i), proj)
            setattr(self, "up_" + str(i), up)
            setattr(self, "node_" + str(i), node)

    def forward(self, layers, startp, endp):
        for i in range(startp + 1, endp):
            upsample = getattr(self, "up_" + str(i - startp))
            project = getattr(self, "proj_" + str(i - startp))
            layers[i] = upsample(project(layers[i]))
            node = getattr(self, "node_" + str(i - startp))
            layers[i] = node(layers[i] + layers[i - 1])


class DLAUp(nn.Module):
    def __init__(self, startp, channels, scales, in_channels=None):
        super(DLAUp, self).__init__()
        self.startp = startp
        if in_channels is None:
            in_channels = channels
        self.channels = channels
        channels = list(channels)
        scales = np.array(scales, dtype=int)
        for i in range(len(channels) - 1):
            j = -i - 2
            setattr(
                self,
                "ida_{}".format(i),
                IDAUp(channels[j], in_channels[j:], scales[j:] // scales[j]),
            )
            scales[j + 1 :] = scales[j]
            in_channels[j + 1 :] = [channels[j] for _ in channels[j + 1 :]]

    def forward(self, layers):
        out = [layers[-1]]  # start with 32
        for i in range(len(layers) - self.startp - 1):
            ida = getattr(self, "ida_{}".format(i))
            ida(layers, len(layers) - i - 2, len(layers))
            out.insert(0, layers[-1])
        return out


class Interpolate(nn.Module):
    def __init__(self, scale, mode):
        super(Interpolate, self).__init__()
        self.scale = scale
        self.mode = mode

    def forward(self, x):
        x = F.interpolate(
            x, scale_factor=self.scale, mode=self.mode, align_corners=False
        )
        return x


class DLASeg(nn.Module):
    def __init__(
        self,
        base_name,
        heads,
        pretrained,
        down_ratio,
        final_kernel,
        last_level,
        head_conv,
        out_channel=0,
        num_classes=None
    ):
        super(DLASeg, self).__init__()
        assert down_ratio in [2, 4, 8, 16]
        self.first_level = int(np.log2(down_ratio))
        self.last_level = last_level
        self.base = globals()[base_name](pretrained=pretrained, num_classes=num_classes)
        channels = self.base.channels
        scales = [2 ** i for i in range(len(channels[self.first_level :]))]
        self.dla_up = DLAUp(self.first_level, channels[self.first_level :], scales)

        if out_channel == 0:
            out_channel = channels[self.first_level]

        self.ida_up = IDAUp(
            out_channel,
            channels[self.first_level : self.last_level],
            [2 ** i for i in range(self.last_level - self.first_level)],
        )

        self.heads = heads
        for head in self.heads:
            classes = self.heads[head]
            if head_conv > 0:
                fc = nn.Sequential(
                    nn.Conv2d(
                        channels[self.first_level],
                        head_conv,
                        kernel_size=3,
                        padding=1,
                        bias=True,
                    ),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(
                        head_conv,
                        classes,
                        kernel_size=final_kernel,
                        stride=1,
                        padding=final_kernel // 2,
                        bias=True,
                    ),
                )
                if "hm" in head:
                    fc[-1].bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            else:
                fc = nn.Conv2d(
                    channels[self.first_level],
                    classes,
                    kernel_size=final_kernel,
                    stride=1,
                    padding=final_kernel // 2,
                    bias=True,
                )
                if "hm" in head:
                    fc.bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            self.__setattr__(head, fc)

    def forward(self, x):
        x = self.base(x)
        x = self.dla_up(x)

        y = []
        for i in range(self.last_level - self.first_level):
            y.append(x[i].clone())
        self.ida_up(y, 0, len(y))

        z = {}
        for head in self.heads:
            z[head] = self.__getattr__(head)(y[-1])
        return [z]


def get_pose_net(pretrained, num_layers, heads, head_conv=256, down_ratio=4, num_classes=None):
    model = DLASeg(
        "dla{}".format(num_layers),
        heads,
        pretrained=pretrained,
        down_ratio=down_ratio,
        final_kernel=1,
        last_level=5,
        head_conv=head_conv,
        num_classes=num_classes,
    )
    return model



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# train function is based on deep_sort/deep/train.py

import os
import torch
import time
from opendr.perception.object_tracking_2d.deep_sort.algorithm.deep_sort.deep.model import Net
from opendr.engine.datasets import MappedDatasetIterator
from opendr.perception.object_tracking_2d.logger import Logger


def train(
    net: Net, dataset, epochs, iters, val_dataset, optimizer,
    batch_size, val_epochs, train_transforms, val_transforms,
    device, log, log_interval, checkpoints_path, checkpoint_after_iter
):

    criterion = torch.nn.CrossEntropyLoss()

    train_loader = torch.utils.data.DataLoader(
        MappedDatasetIterator(dataset, lambda data: (train_transforms(data[0]), data[1])),
        batch_size=batch_size,
        shuffle=True,
    )
    val_loader = torch.utils.data.DataLoader(
        MappedDatasetIterator(val_dataset, lambda data: (val_transforms(data[0]), data[1])),
        batch_size=batch_size,
        shuffle=True,
    )

    training_loss = 0.0
    train_loss = 0.0
    correct = 0
    total = 0
    start = time.time()
    last_val_loss = -1

    if iters == -1:
        iters = len(train_loader)

    iter = 0

    for epoch in range(epochs):

        net.train()

        for idx, (inputs, labels) in enumerate(train_loader):

            if idx > iters:
                continue

            # forward
            inputs, labels = inputs.to(device), labels.to(device)
            outputs = net(inputs)
            loss = criterion(outputs, labels)

            # backward
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            # accumurating
            training_loss += loss.item()
            train_loss += loss.item()
            correct += outputs.max(dim=1)[1].eq(labels).sum().item()
            total += labels.size(0)

            # print
            if (idx + 1) % log_interval == 0:
                end = time.time()
                log(
                    Logger.LOG_WHEN_NORMAL,
                    "[{}/{}][progress:{:.1f}%]time:{:.2f}s Loss:{:.5f} Correct:{}/{} Acc:{:.3f}%".format(
                        epoch + 1,
                        epochs,
                        100.0 * (idx + 1) / len(train_loader),
                        end - start,
                        training_loss / log_interval,
                        correct,
                        total,
                        100.0 * correct / total,
                    )
                )
                training_loss = 0.0
                start = time.time()

            if checkpoint_after_iter > 0:
                if ((iter + 1) % checkpoint_after_iter == 0):
                    torch.save(net.state_dict(), os.path.join(
                        checkpoints_path,
                        f"checkpoint_{iter}.pth"
                    ))

        if val_epochs > 0 and (epoch + 1) % val_epochs == 0:
            log(Logger.LOG_WHEN_NORMAL, "Eval")
            net.eval()
            test_loss = 0.0
            correct = 0
            total = 0
            start = time.time()
            with torch.no_grad():
                for idx, (inputs, labels) in enumerate(val_loader):

                    inputs, labels = inputs.to(device), labels.to(device)
                    outputs = net(inputs)
                    loss = criterion(outputs, labels)

                    test_loss += loss.item()
                    correct += outputs.max(dim=1)[1].eq(labels).sum().item()
                    total += labels.size(0)

                end = time.time()
                log(
                    Logger.LOG_WHEN_NORMAL,
                    "[progress:{:.1f}%]time:{:.2f}s Loss:{:.5f} Correct:{}/{} Acc:{:.3f}%".format(
                        100.0 * (idx + 1) / len(val_loader),
                        end - start,
                        test_loss / len(val_loader),
                        correct,
                        total,
                        100.0 * correct / total,
                    )
                )

            last_val_loss = test_loss

    result = {"last_val_loss": last_val_loss}

    return result



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# general imports
import os
import time
import json
import numpy as np
from multiprocessing import Pool
import cv2
from tqdm import tqdm
from urllib.request import urlretrieve

# gluoncv imports
import mxnet as mx
from mxnet import gluon, nd, autograd
from gluoncv import utils as gutils
from gluoncv import model_zoo
from gluoncv.model_zoo.siamrpn.siamrpn_tracker import SiamRPNTracker as build_tracker
from gluoncv.loss import SiamRPNLoss
from gluoncv.utils import LRScheduler, LRSequential, split_and_load
from gluoncv.model_zoo.siamrpn.siamrpn_tracker import get_axis_aligned_bbox
from gluoncv.data.tracking_data.track import TrkDataset
from gluoncv.utils.metrics.tracking import OPEBenchmark
from gluoncv.data.otb.tracking import OTBTracking

# OpenDR engine imports
from opendr.engine.learners import Learner
from opendr.engine.data import Image
from opendr.engine.target import TrackingAnnotation
from opendr.engine.datasets import ExternalDataset
from opendr.engine.constants import OPENDR_SERVER_URL
from opendr.engine.datasets import DatasetIterator

gutils.random.seed(0)
os.environ['MXNET_CUDNN_AUTOTUNE_DEFAULT'] = '0'


class SiamRPNLearner(Learner):
    def __init__(self, device='cuda', n_epochs=50, num_workers=1, warmup_epochs=2,
                 lr=1e-3, weight_decay=0, momentum=0.9, cls_weight=1., loc_weight=1.2,
                 batch_size=32, temp_path=''):
        """
        SiamRPN Tracker Learner
        :param device: Either 'cpu' or 'cuda'. If a specific GPU is to be used, can be of the form 'cuda:#'
        :type device: str, optional
        :param n_epochs: Total number of epochs to train for
        :type n_epochs: int, optional
        :param num_workers: Number of threads used to load the train dataset or perform evaluation
        :type num_workers: int, optional
        :param warmup_epochs: Number of epochs during which the learning rate is annealer to `lr`
        :type warmup_epochs: int, optional
        :param lr: Initial learning rate, after warmup_epochs
        :type lr: float, optional
        :param weight_decay: Weight decay factor
        :type weight_decay: float, optional
        :param momentum: Optimizer momentum
        :type momentum: float, optional
        :param cls_weight: Weighs the classification loss
        :type cls_weight: float, optional
        :param loc_weight: Weights the localization loss
        :type loc_weight: float, optional
        :param batch_size: Batch size for training
        :type batch_size: int, optional
        :param temp_path: path to where relevant data and weights are downloaded
        :type temp_path: str, optional
        """
        self.loc_weight = loc_weight
        self.cls_weight = cls_weight
        self.warmup_epochs = warmup_epochs
        self.num_workers = num_workers
        self.n_epochs = n_epochs
        backbone = 'siamrpn_alexnet_v2_otb15'
        super(SiamRPNLearner, self).__init__(device=device, backbone=backbone, lr=lr,
                                             batch_size=batch_size, temp_path=temp_path)
        self.weight_decay = weight_decay
        self.momentum = momentum

        if 'cuda' in self.device:
            if mx.context.num_gpus() > 0:
                if self.device == 'cuda':
                    self.ctx = mx.gpu(0)
                else:
                    self.ctx = mx.gpu(int(self.device.split(':')[1]))
            else:
                self.ctx = mx.cpu()
        else:
            self.ctx = mx.cpu()

        self.__create_model()
        self.tracker = build_tracker(self._model)

    def __create_model(self, pretrained=True):
        """base model creation"""
        self._model = model_zoo.get_model(self.backbone, ctx=self.ctx, pretrained=pretrained)

    def fit(self, dataset, log_interval=20, n_gpus=1,
            val_dataset=None, logging_path='', silent=True, verbose=True):
        """
        Train the tracker on a new dataset.
        :param dataset: training dataset
        :type dataset: ExternalDataset or supported DatasetIterator
        :param log_interval: Train loss is printed after log_interval iterations
        :type log_interval: int
        :param n_gpus: Number of GPUs to train with if device is set to GPU
        :type n_gpus: int
        :param verbose: if set to True, additional information is printed to STDOUT, defaults to True
        :type verbose: bool
        :param val_dataset: ignored
        :type val_dataset: ExternalDataset, optional
        :param logging_path: ignored
        :type logging_path: str, optional
        :param silent: ignored
        :type silent: str, optional
        :param verbose: if set to True, additional information is printed to STDOUT, defaults to True
        :type verbose: bool
        :return: returns stats regarding the training process
        :rtype: dict
        """
        dataset = self.__prepare_training_dataset(dataset)
        train_loader = gluon.data.DataLoader(dataset,
                                             batch_size=self.batch_size,
                                             last_batch='discard',
                                             num_workers=self.num_workers)

        if self.device.startswith('cuda'):
            if ':' in self.device:
                _, gpu_no = self.device.split(':')
                ctx = [mx.gpu(int(gpu_no))]
            else:
                ctx = [mx.gpu(i) for i in range(n_gpus)]
        else:
            ctx = [mx.cpu(0)]

        self._model = model_zoo.get_model(self.backbone, bz=self.batch_size, is_train=True, ctx=ctx,
                                          pretrained=True)

        criterion = SiamRPNLoss(self.batch_size)
        step_epoch = [10 * i for i in range(0, self.n_epochs, 10)]
        num_batches = len(train_loader)
        lr_scheduler = LRSequential([LRScheduler(mode='step',
                                                 base_lr=0.005,
                                                 target_lr=0.01,
                                                 nepochs=self.warmup_epochs,
                                                 iters_per_epoch=num_batches,
                                                 step_epoch=step_epoch,
                                                 ),
                                     LRScheduler(mode='poly',
                                                 base_lr=0.01,
                                                 target_lr=0.005,
                                                 nepochs=self.n_epochs - self.warmup_epochs,
                                                 iters_per_epoch=num_batches,
                                                 step_epoch=[e - self.warmup_epochs for e in step_epoch],
                                                 power=0.02)])

        optimizer_params = {'lr_scheduler': lr_scheduler,
                            'wd': self.weight_decay,
                            'momentum': self.momentum,
                            'learning_rate': self.lr}
        optimizer = gluon.Trainer(self._model.collect_params(), 'sgd', optimizer_params)
        train_dict = {
            'loss_total': [],
            'loss_loc': [],
            'loss_cls': []
        }

        for epoch in range(self.n_epochs):
            loss_total_val = 0
            loss_loc_val = 0
            loss_cls_val = 0
            batch_time = time.time()
            for i, data in enumerate(train_loader):
                template, search, label_cls, label_loc, label_loc_weight = self.__train_batch_fn(data, ctx)
                cls_losses = []
                loc_losses = []
                total_losses = []

                with autograd.record():
                    for j in range(len(ctx)):
                        cls, loc = self._model(template[j], search[j])
                        label_cls_temp = label_cls[j].reshape(-1).asnumpy()
                        pos_index = np.argwhere(label_cls_temp == 1).reshape(-1)
                        neg_index = np.argwhere(label_cls_temp == 0).reshape(-1)
                        if len(pos_index):
                            pos_index = nd.array(pos_index, ctx=ctx[j])
                        else:
                            pos_index = nd.array(np.array([]), ctx=ctx[j])
                        if len(neg_index):
                            neg_index = nd.array(neg_index, ctx=ctx[j])
                        else:
                            neg_index = nd.array(np.array([]), ctx=ctx[j])
                        cls_loss, loc_loss = criterion(cls, loc, label_cls[j], pos_index, neg_index,
                                                       label_loc[j], label_loc_weight[j])
                        total_loss = self.cls_weight * cls_loss + self.loc_weight * loc_loss
                        cls_losses.append(cls_loss)
                        loc_losses.append(loc_loss)
                        total_losses.append(total_loss)

                    mx.nd.waitall()
                    autograd.backward(total_losses)
                optimizer.step(self.batch_size)
                loss_total_val += sum([l.mean().asscalar() for l in total_losses]) / len(total_losses)
                loss_loc_val += sum([l.mean().asscalar() for l in loc_losses]) / len(loc_losses)
                loss_cls_val += sum([l.mean().asscalar() for l in cls_losses]) / len(cls_losses)
                if i % log_interval == 0 and verbose:
                    print('Epoch %d iteration %04d/%04d: loc loss %.3f, cls loss %.3f, \
                                     training loss %.3f, batch time %.3f' %
                          (epoch, i, len(train_loader), loss_loc_val / (i + 1), loss_cls_val / (i + 1),
                           loss_total_val / (i + 1), time.time() - batch_time))
                    batch_time = time.time()
                mx.nd.waitall()
            train_dict['loss_total'].append(loss_total_val)
            train_dict['loss_loc'].append(loss_loc_val)
            train_dict['loss_cls'].append(loss_cls_val)
        return train_dict

    def eval(self, dataset):
        """
        Evaluate the current model on the OTB dataset. Measures success and FPS.
        :param dataset: Dataset for evaluation.
        :type dataset: `ExternalDataset`
        :return: returns stats regarding evaluation
        :rtype: dict
        """
        tracker_name = self.backbone

        dataset = self.__prepare_validation_dataset(dataset)

        self._model.collect_params().reset_ctx(self.ctx)
        self.tracker = build_tracker(self._model)

        # iterate through dataset
        fps = np.zeros(len(dataset))
        for v_idx, video in enumerate(dataset):
            toc = 0
            pred_bboxes = []
            scores = []
            track_times = []
            for idx, (img, gt_bbox) in enumerate(video):
                tic = cv2.getTickCount()
                if idx == 0:
                    x_max, y_max, gt_w, gt_t = get_axis_aligned_bbox(np.array(gt_bbox))
                    gt_bbox_ = [x_max - (gt_w - 1) / 2, y_max - (gt_t - 1) / 2, gt_w, gt_t]
                    self.tracker.init(img, gt_bbox_, self.ctx)
                    pred_bbox = gt_bbox_
                    scores.append(None)
                    pred_bboxes.append(pred_bbox)
                else:
                    outputs = self.tracker.track(img, self.ctx)
                    pred_bbox = outputs['bbox']
                    pred_bboxes.append(pred_bbox)
                    scores.append(outputs['best_score'])
                toc += cv2.getTickCount() - tic
                track_times.append((cv2.getTickCount() - tic) / cv2.getTickFrequency())
            toc /= cv2.getTickFrequency()

            print('({:3d}) Video: {:12s} Time: {:5.1f}s Speed: {:3.1f}fps'.format
                  (v_idx + 1, video.name, toc, len(video) / toc))
            video.pred_trajs[tracker_name] = pred_bboxes
            fps[v_idx] = len(video) / toc

        mean_fps = np.mean(fps)
        benchmark = OPEBenchmark(dataset)
        trackers = [tracker_name]
        success_ret = {}
        with Pool(processes=self.num_workers) as pool:
            for ret in tqdm(pool.imap_unordered(benchmark.eval_success, trackers),
                            desc='eval success', total=len(trackers), ncols=100):
                success_ret.update(ret)
        precision_ret = {}
        with Pool(processes=self.num_workers) as pool:
            for ret in tqdm(pool.imap_unordered(benchmark.eval_precision, trackers),
                            desc='eval precision', total=len(trackers), ncols=100):
                precision_ret.update(ret)

        tracker_auc = {}
        for tracker_name in success_ret.keys():
            auc = np.mean(list(success_ret[tracker_name].values()))
            tracker_auc[tracker_name] = auc

        tracker_name_len = max((max([len(x) for x in success_ret.keys()]) + 2), 12)
        header = ("|{:^" + str(tracker_name_len) + "}|{:^9}|{:^9}|").format("Tracker name", "Success", "FPS")
        formatter = "|{:^" + str(tracker_name_len) + "}|{:^9.3f}|{:^9.1f}|"
        print('-' * len(header))
        print(header)
        print('-' * len(header))
        success = tracker_auc[tracker_name]
        print(formatter.format(tracker_name, success, mean_fps))
        print('-' * len(header))
        eval_dict = {
            'success': success,
            'fps': mean_fps
        }
        return eval_dict

    def infer(self, img, init_box=None):
        """
        Performs inference on an input image and returns the resulting bounding box.
        :param img: image to perform inference on
        :type img: opendr.engine.data.Image
        :param init_box: If provided, it is used to initialized the tracker on the contained object
        :type init_box: TrackingAnnotation
        :return: list of bounding boxes
        :rtype: BoundingBoxList
        """
        if isinstance(img, Image):
            img = img.opencv()

        if isinstance(init_box, TrackingAnnotation) and init_box is not None:
            # initialize tracker
            gt_bbox_ = [init_box.left, init_box.top, init_box.width, init_box.height]
            self.tracker.init(img, gt_bbox_, ctx=self.ctx)
            pred_bbox = gt_bbox_
        else:
            outputs = self.tracker.track(img, ctx=self.ctx)
            pred_bbox = outputs['bbox']

        pred_bbox = list(map(int, pred_bbox))
        return TrackingAnnotation(left=pred_bbox[0], top=pred_bbox[1],
                                  width=pred_bbox[2], height=pred_bbox[3], name=0, id=0)

    def save(self, path, verbose=False):
        """
        Method for saving the current model in the path provided.
        :param path: path to folder where model will be saved
        :type path: str
        :param verbose: whether to print a success message or not, defaults to False
        :type verbose: bool, optional
        """
        os.makedirs(path, exist_ok=True)

        model_name = os.path.basename(path)
        if verbose:
            print(model_name)
        metadata = {"model_paths": [], "framework": "mxnet", "format": "params",
                    "has_data": False, "inference_params": {}, "optimized": False,
                    "optimizer_info": {}, "backbone": self.backbone}
        param_filepath = model_name + ".params"
        metadata["model_paths"].append(param_filepath)

        self._model.save_parameters(os.path.join(path, metadata["model_paths"][0]))
        if verbose:
            print("Model parameters saved.")

        with open(os.path.join(path, model_name + '.json'), 'w', encoding='utf-8') as f:
            json.dump(metadata, f, ensure_ascii=False, indent=4)
        if verbose:
            print("Model metadata saved.")
        return True

    def load(self, path, verbose=False):
        """
        Loads the model from the path provided, based on the metadata .json file included.
        :param path: path of the directory where the model was saved
        :type path: str
        :param verbose: whether to print a success message or not, defaults to False
        :type verbose: bool, optional
        """
        model_name = os.path.basename(os.path.normpath(path))
        if verbose:
            print("Model name:", model_name, "-->", os.path.join(path, model_name + ".json"))
        with open(os.path.join(path, model_name + ".json")) as f:
            metadata = json.load(f)

        self.backbone = metadata["backbone"]
        self.__create_model(pretrained=False)

        self._model.load_parameters(os.path.join(path, metadata["model_paths"][0]))
        self._model.collect_params().reset_ctx(self.ctx)
        self._model.hybridize(static_alloc=True, static_shape=True)
        if verbose:
            print("Loaded parameters and metadata.")
        return True

    def download(self, path=None, mode="pretrained", verbose=False,
                 url=OPENDR_SERVER_URL + "/perception/object_tracking_2d/siamrpn/",
                 overwrite=False):
        """
        Downloads all files necessary for inference, evaluation and training. Valid mode options are: ["pretrained",
        "video", "test_data", "otb2015"].
        :param path: folder to which files will be downloaded, if None self.temp_path will be used
        :type path: str, optional
        :param mode: one of: ["pretrained", "video", "test_data", "otb2015"], where "pretrained" downloads a pretrained
        network, "video" downloads example inference data, "test_data" downloads a very small train/eval subset, and
         "otb2015" downloads the OTB dataset
        :type mode: str, optional
        :param verbose: if True, additional information is printed on stdout
        :type verbose: bool, optional
        :param url: URL to file location on FTP server
        :type url: str, optional
        :param overwrite: if True, the downloaded files will be overwritten
        :type overwrite:bool, optional
        """
        valid_modes = ["pretrained", "video", "otb2015", "test_data"]
        if mode not in valid_modes:
            raise UserWarning("mode parameter not valid:", mode, ", file should be one of:", valid_modes)

        if path is None:
            path = self.temp_path

        if not os.path.exists(path):
            os.makedirs(path)

        if mode == "pretrained":
            path = os.path.join(path, "siamrpn_opendr")
            if not os.path.exists(path):
                os.makedirs(path)

            if verbose:
                print("Downloading pretrained model...")

            file_url = os.path.join(url, "pretrained",
                                    "siamrpn_opendr",
                                    "siamrpn_opendr.json")
            if verbose:
                print("Downloading metadata...")
            file_path = os.path.join(path, "siamrpn_opendr.json")
            if not os.path.exists(file_path) or overwrite:
                urlretrieve(file_url, file_path)

            if verbose:
                print("Downloading params...")
            file_url = os.path.join(url, "pretrained", "siamrpn_opendr",
                                    "siamrpn_opendr.params")
            file_path = os.path.join(path, "siamrpn_opendr.params")
            if not os.path.exists(file_path) or overwrite:
                urlretrieve(file_url, file_path)

        elif mode == "video":
            file_url = os.path.join(url, "video", "tc_Skiing_ce.mp4")
            if verbose:
                print("Downloading example video...")
            file_path = os.path.join(path, "tc_Skiing_ce.mp4")
            if not os.path.exists(file_path) or overwrite:
                urlretrieve(file_url, file_path)

        elif mode == "test_data":
            os.makedirs(os.path.join(path, "Basketball"), exist_ok=True)
            os.makedirs(os.path.join(path, "Basketball", "img"), exist_ok=True)
            # download annotation
            file_url = os.path.join(url, "test_data", "OTBtest.json")
            if verbose:
                print("Downloading annotation...")
            file_path = os.path.join(path, "OTBtest.json")
            if not os.path.exists(file_path) or overwrite:
                urlretrieve(file_url, file_path)
            # download image
            if verbose:
                print("Downloading 100 images...")
            for i in range(100):
                file_url = os.path.join(url, "test_data", "Basketball", "img", f"{i+1:04d}.jpg")
                file_path = os.path.join(path, "Basketball", "img", f"{i+1:04d}.jpg")
                if not os.path.exists(file_path) or overwrite:
                    urlretrieve(file_url, file_path)

        else:
            # mode == 'otb2015'
            from .data_utils.otb import download_otb
            if verbose:
                print('Attempting to download OTB2015 (100 videos)...')
            download_otb(os.path.join(path, "otb2015"), overwrite=overwrite)
            file_url = os.path.join(url, "otb2015", "OTB2015.json")
            if verbose:
                print("Downloading annotation...")
            file_path = os.path.join(path, "otb2015", "OTB2015.json")
            if not os.path.exists(file_path) or overwrite:
                urlretrieve(file_url, file_path)

    @staticmethod
    def __train_batch_fn(data, ctx):
        """split and load data in GPU"""
        template = split_and_load(data[0], ctx_list=ctx, batch_axis=0)
        search = split_and_load(data[1], ctx_list=ctx, batch_axis=0)
        label_cls = split_and_load(data[2], ctx_list=ctx, batch_axis=0)
        label_loc = split_and_load(data[3], ctx_list=ctx, batch_axis=0)
        label_loc_weight = split_and_load(data[4], ctx_list=ctx, batch_axis=0)
        return template, search, label_cls, label_loc, label_loc_weight

    def __prepare_training_dataset(self, dataset):
        """
        Converts `ExternalDataset` or list of `ExternalDatasets` to appropriate format.
        :param dataset: Training dataset(s)
        :type dataset: `ExternalDataset` or list of `ExternalDatasets`
        """
        if isinstance(dataset, list) or isinstance(dataset, tuple):
            frame_range_map = {
                'vid': 100,
                'Youtube_bb': 3,
                'coco': 1,
                'det': 1,
            }
            num_use_map = {
                'vid': 100000,
                'Youtube_bb': -1,
                'coco': -1,
                'det': -1,
            }

            dataset_paths = []
            dataset_names = []
            dataset_roots = []
            dataset_annos = []
            frame_ranges = []
            num_uses = []

            for _dataset in dataset:
                # check if all are ExternalDataset
                if not isinstance(dataset, ExternalDataset):
                    raise TypeError("Only `ExternalDataset` types are supported.")
                # get params
                dataset_paths.append(_dataset.path)
                dataset_names.append(_dataset.dataset_type)
                dataset_roots.append(os.path.join(_dataset.dataset_type, 'crop511'))
                dataset_annos.append(os.path.join(_dataset.dataset_type,
                                                  f'train{"2017" if _dataset.dataset_type == "coco" else ""}.json'))
                frame_ranges.append(frame_range_map[_dataset.dataset_type])
                num_uses.append(num_use_map[_dataset.dataset_type])
            dataset = TrkDataset(dataset[0].path,
                                 dataset_names=dataset_names, detaset_root=dataset_roots,
                                 detaset_anno=dataset_annos, train_epoch=self.n_epochs,
                                 dataset_frame_range=frame_ranges, dataset_num_use=num_uses)
            return dataset

        if isinstance(dataset, ExternalDataset):
            dataset_types = ['vid', 'Youtube_bb', 'coco', 'det']
            assert dataset.dataset_type in dataset_types, f"Unrecognized dataset_type," \
                                                          f" acceptable values: {dataset_types}"
            dataset = TrkDataset(data_path=dataset.path,
                                 dataset_names=[dataset.dataset_type], detaset_root=[f'{dataset.dataset_type}/crop511'],
                                 detaset_anno=[f'{dataset.dataset_type}/'
                                               f'train{"2017" if dataset.dataset_type == "coco" else ""}.json'],
                                 train_epoch=self.n_epochs)
            return dataset

        if issubclass(type(dataset), DatasetIterator):
            return dataset

        if not isinstance(dataset, ExternalDataset):
            raise TypeError("Only `ExternalDataset` and modified `DatasetIterator` types are supported.")

    @staticmethod
    def __prepare_validation_dataset(dataset):
        """
        :param dataset: `ExternalDataset` object containing OTB2015 dataset root and type ('OTB2015')
        :type dataset: ExternalDataset
        """
        if not isinstance(dataset, ExternalDataset):
            raise TypeError("Only `ExternalDataset` types are supported.")
        dataset_types = ["OTB2015", "OTBtest"]
        assert dataset.dataset_type in dataset_types, "Unrecognized dataset type, only OTB2015 is supported currently"
        dataset = OTBTracking(dataset.dataset_type, dataset_root=dataset.path, load_img=False)
        return dataset

    def reset(self):
        """This method is not used in this implementation."""
        return NotImplementedError

    def optimize(self, target_device):
        """This method is not used in this implementation."""
        return NotImplementedError



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from .image import transform_preds


def ctdet_post_process(dets, c, s, h, w, num_classes):
    # dets: batch x max_dets x dim
    # return 1-based class det dict
    ret = []
    for i in range(dets.shape[0]):
        top_preds = {}
        dets[i, :, :2] = transform_preds(dets[i, :, 0:2], c[i], s[i], (w, h))
        dets[i, :, 2:4] = transform_preds(dets[i, :, 2:4], c[i], s[i], (w, h))
        classes = dets[i, :, -1]
        for j in range(num_classes):
            inds = classes == j
            top_preds[j + 1] = np.concatenate(
                [
                    dets[i, inds, :4].astype(np.float32),
                    dets[i, inds, 4:5].astype(np.float32),
                ],
                axis=1,
            ).tolist()
        ret.append(top_preds)
    return ret



import logging


def get_logger(name="root"):
    formatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s]: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger


logger = get_logger("root")



# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Bin Xiao (Bin.Xiao@microsoft.com)
# Modified by Dequan Wang and Xingyi Zhou
# ------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import logging

import torch
import torch.nn as nn
from DCN import DCN
import torch.utils.model_zoo as model_zoo

BN_MOMENTUM = 0.1
logger = logging.getLogger(__name__)

model_urls = {
    "resnet18": "https://download.pytorch.org/models/resnet18-5c106cde.pth",
    "resnet34": "https://download.pytorch.org/models/resnet34-333f7ec4.pth",
    "resnet50": "https://download.pytorch.org/models/resnet50-19c8e357.pth",
    "resnet101": "https://download.pytorch.org/models/resnet101-5d3b4d8f.pth",
    "resnet152": "https://download.pytorch.org/models/resnet152-b121ed2d.pth",
}


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=stride, padding=1, bias=False
        )
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(
            planes, planes * self.expansion, kernel_size=1, bias=False
        )
        self.bn3 = nn.BatchNorm2d(planes * self.expansion, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


def fill_up_weights(up):
    w = up.weight.data
    f = math.ceil(w.size(2) / 2)
    c = (2 * f - 1 - f % 2) / (2.0 * f)
    for i in range(w.size(2)):
        for j in range(w.size(3)):
            w[0, 0, i, j] = (1 - math.fabs(i / f - c)) * (1 - math.fabs(j / f - c))
    for c in range(1, w.size(0)):
        w[c, 0, :, :] = w[0, 0, :, :]


def fill_fc_weights(layers):
    for m in layers.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.normal_(m.weight, std=0.001)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)


class PoseResNet(nn.Module):
    def __init__(self, block, layers, heads, head_conv):
        self.inplanes = 64
        self.heads = heads
        self.deconv_with_bias = False

        super(PoseResNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)

        # used for deconv layers
        self.deconv_layers = self._make_deconv_layer(3, [256, 128, 64], [4, 4, 4],)

        for head in self.heads:
            classes = self.heads[head]
            if head_conv > 0:
                fc = nn.Sequential(
                    nn.Conv2d(64, head_conv, kernel_size=3, padding=1, bias=True),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(
                        head_conv,
                        classes,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        bias=True,
                    ),
                )
                if "hm" in head:
                    fc[-1].bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            else:
                fc = nn.Conv2d(
                    64, classes, kernel_size=1, stride=1, padding=0, bias=True
                )
                if "hm" in head:
                    fc.bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            self.__setattr__(head, fc)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(planes * block.expansion, momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def _get_deconv_cfg(self, deconv_kernel, index):
        if deconv_kernel == 4:
            padding = 1
            output_padding = 0
        elif deconv_kernel == 3:
            padding = 1
            output_padding = 1
        elif deconv_kernel == 2:
            padding = 0
            output_padding = 0

        return deconv_kernel, padding, output_padding

    def _make_deconv_layer(self, num_layers, num_filters, num_kernels):
        assert num_layers == len(
            num_filters
        ), "ERROR: num_deconv_layers is different len(num_deconv_filters)"
        assert num_layers == len(
            num_kernels
        ), "ERROR: num_deconv_layers is different len(num_deconv_filters)"

        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = self._get_deconv_cfg(num_kernels[i], i)

            planes = num_filters[i]
            fc = DCN(
                self.inplanes,
                planes,
                kernel_size=(3, 3),
                stride=1,
                padding=1,
                dilation=1,
                deformable_groups=1,
            )

            up = nn.ConvTranspose2d(
                in_channels=planes,
                out_channels=planes,
                kernel_size=kernel,
                stride=2,
                padding=padding,
                output_padding=output_padding,
                bias=self.deconv_with_bias,
            )
            fill_up_weights(up)

            layers.append(fc)
            layers.append(nn.BatchNorm2d(planes, momentum=BN_MOMENTUM))
            layers.append(nn.ReLU(inplace=True))
            layers.append(up)
            layers.append(nn.BatchNorm2d(planes, momentum=BN_MOMENTUM))
            layers.append(nn.ReLU(inplace=True))
            self.inplanes = planes

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.deconv_layers(x)
        ret = {}
        for head in self.heads:
            ret[head] = self.__getattr__(head)(x)
        return [ret]

    def init_weights(self, pretrained, num_layers):
        if pretrained:
            url = model_urls["resnet{}".format(num_layers)]
            pretrained_state_dict = model_zoo.load_url(url)
            print("=> loading pretrained model {}".format(url))
            self.load_state_dict(pretrained_state_dict, strict=False)
        print("=> init deconv weights from normal distribution")
        for name, m in self.deconv_layers.named_modules():
            if isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


resnet_spec = {
    18: (BasicBlock, [2, 2, 2, 2]),
    34: (BasicBlock, [3, 4, 6, 3]),
    50: (Bottleneck, [3, 4, 6, 3]),
    101: (Bottleneck, [3, 4, 23, 3]),
    152: (Bottleneck, [3, 8, 36, 3]),
}


def get_pose_net(pretrained, num_layers, heads, head_conv=256):
    block_class, layers = resnet_spec[num_layers]

    model = PoseResNet(block_class, layers, heads, head_conv=head_conv)
    model.init_weights(pretrained, num_layers)
    return model
